# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2013 WandererFan <wandererfan@gmail.com>                *
# *   Copyright (c) 2019 Zheng, Lei (realthunder)<realthunder.dev@gmail.com>*
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create PathArray objects.

The copies will be placed along a path like a polyline, spline, or bezier
curve.
"""
## @package make_patharray
# \ingroup draftmake
# \brief Provides functions to create PathArray objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftutils.messages import _err
from draftutils.translate import translate
from draftobjects.patharray import PathArray
from draftobjects.pathtwistedarray import PathTwistedArray

if App.GuiUp:
    from draftviewproviders.view_array import ViewProviderDraftArray
    from draftviewproviders.view_draftlink import ViewProviderDraftLink


def make_path_array(
    base_object,
    path_object,
    count=4,
    extra=App.Vector(0, 0, 0),
    subelements=None,
    align=False,
    align_mode="Original",
    tan_vector=App.Vector(1, 0, 0),
    force_vertical=False,
    vertical_vector=App.Vector(0, 0, 1),
    start_offset=0.0,
    end_offset=0.0,
    use_link=True,
):
    """Make a Draft PathArray object.

    Distribute copies of a `base_object` along `path_object`
    or `subelements` from `path_object`.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    path_object: Part::Feature or str
        Path object like a polyline, B-Spline, or bezier curve that should
        contain edges.
        Just like `base_object` it can also be `Label`.

    count: int, float, optional
        It defaults to 4.
        Number of copies to create along the `path_object`.
        It must be at least 2.
        If a `float` is provided, it will be truncated by `int(count)`.

    extra: Base.Vector3, optional
        It defaults to `App.Vector(0, 0, 0)`.
        It translates each copy by the value of `extra`.
        This is useful to adjust for the difference between shape centre
        and shape reference point.

    subelements: list or tuple of str, optional
        It defaults to `None`.
        It should be a list of names of edges that must exist in `path_object`.
        Then the path array will be created along these edges only,
        and not the entire `path_object`.
        ::
            subelements = ['Edge1', 'Edge2']

        The edges must be contiguous, meaning that it is not allowed to
        input `'Edge1'` and `'Edge3'` if they do not touch each other.

        A single string value is also allowed.
        ::
            subelements = 'Edge1'

    align: bool, optional
        It defaults to `False`.
        If it is `True` it will align `base_object` to tangent, normal,
        or binormal to the `path_object`, depending on the value
        of `tan_vector`.

    align_mode: str, optional
        It defaults to `'Original'` which is the traditional alignment.
        It can also be `'Frenet'` or `'Tangent'`.

        - Original. It does not calculate curve normal.
          `X` is curve tangent, `Y` is normal parameter, Z is the cross
          product `X` x `Y`.
        - Frenet. It defines a local coordinate system along the path.
          `X` is tangent to curve, `Y` is curve normal, `Z` is curve binormal.
          If normal cannot be computed, for example, in a straight path,
          a default is used.
        - Tangent. It is similar to `'Original'` but includes a pre-rotation
          to align the base object's `X` to the value of `tan_vector`,
          then `X` follows curve tangent.

    tan_vector: Base::Vector3, optional
        It defaults to `App.Vector(1, 0, 0)` or the +X axis.
        It aligns the tangent of the path to this local unit vector
        of the object.

    force_vertical: Base::Vector3, optional
        It defaults to `False`.
        If it is `True`, the value of `vertical_vector`
        will be used when `align_mode` is `'Original'` or `'Tangent'`.

    vertical_vector: Base::Vector3, optional
        It defaults to `App.Vector(0, 0, 1)` or the +Z axis.
        It will force this vector to be the vertical direction
        when `force_vertical` is `True`.

    start_offset: float, optional
        It defaults to 0.0.
        It is the length from the start of the path to the first copy.

    end_offset: float, optional
        It defaults to 0.0.
        It is the length from the end of the path to the last copy.

    use_link: bool, optional
        It defaults to `True`, in which case the copies are `App::Link`
        elements. Otherwise, the copies are shape copies which makes
        the resulting array heavier.

    Returns
    -------
    Part::FeaturePython
        The scripted object of type `'PathArray'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.
    """
    _name = "make_path_array"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    found, base_object = utils.find_object(base_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: base_object not in document."))
        return None

    found, path_object = utils.find_object(path_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: path_object not in document."))
        return None

    try:
        utils.type_check([(count, (int, float))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number."))
        return None
    count = int(count)

    try:
        utils.type_check([(extra, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    if subelements:
        try:
            # Make a list
            if isinstance(subelements, str):
                subelements = [subelements]

            utils.type_check([(subelements, (list, tuple, str))], name=_name)
        except TypeError:
            _err(
                translate(
                    "draft", "Wrong input: must be a list or tuple of strings, or a single string."
                )
            )
            return None

        # The subelements list is used to build a special list
        # called a LinkSubList, which includes the path_object.
        # Old style: [(path_object, "Edge1"), (path_object, "Edge2")]
        # New style: [(path_object, ("Edge1", "Edge2"))]
        #
        # If a simple list is given ["a", "b"], this will create an old-style
        # SubList.
        # If a nested list is given [["a", "b"]], this will create a new-style
        # SubList.
        # In any case, the property of the object accepts both styles.
        #
        # If the old style is deprecated then this code should be updated
        # to create new style lists exclusively.
        sub_list = list()
        for sub in subelements:
            sub_list.append((path_object, sub))
    else:
        sub_list = None

    align = bool(align)

    try:
        utils.type_check([(align_mode, str)], name=_name)

        if align_mode not in ("Original", "Frenet", "Tangent"):
            raise TypeError
    except TypeError:
        _err(translate("draft", "Wrong input: must be 'Original', 'Frenet', or 'Tangent'."))
        return None

    try:
        utils.type_check([(tan_vector, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    force_vertical = bool(force_vertical)
    try:
        utils.type_check([(vertical_vector, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    try:
        utils.type_check([(start_offset, (int, float))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number."))
        return None
    start_offset = float(start_offset)

    try:
        utils.type_check([(end_offset, (int, float))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number."))
        return None
    end_offset = float(end_offset)

    use_link = bool(use_link)

    if use_link:
        # The PathArray class must be called in this special way
        # to make it a PathLinkArray
        new_obj = doc.addObject("Part::FeaturePython", "PathArray", PathArray(None), None, True)
    else:
        new_obj = doc.addObject("Part::FeaturePython", "PathArray")
        PathArray(new_obj)

    new_obj.Base = base_object
    new_obj.PathObject = path_object
    new_obj.Count = count
    new_obj.ExtraTranslation = extra
    new_obj.PathSubelements = sub_list
    new_obj.Align = align
    new_obj.AlignMode = align_mode
    new_obj.TangentVector = tan_vector
    new_obj.ForceVertical = force_vertical
    new_obj.VerticalVector = vertical_vector
    new_obj.StartOffset = start_offset
    new_obj.EndOffset = end_offset

    if App.GuiUp:
        if use_link:
            ViewProviderDraftLink(new_obj.ViewObject)
        else:
            ViewProviderDraftArray(new_obj.ViewObject)
            gui_utils.formatObject(new_obj, new_obj.Base)
            new_obj.ViewObject.Proxy.resetColors(new_obj.ViewObject)
        new_obj.Base.ViewObject.hide()
        gui_utils.select(new_obj)

    return new_obj


def makePathArray(
    baseobject, pathobject, count, xlate=None, align=False, pathobjsubs=[], use_link=False
):
    """Create PathArray. DEPRECATED. Use 'make_path_array'."""
    utils.use_instead("make_path_array")

    return make_path_array(baseobject, pathobject, count, xlate, pathobjsubs, align, use_link)


def make_path_twisted_array(base_object, path_object, count=15, rot_factor=0.25, use_link=True):
    """Create a Path twisted array."""
    _name = "make_path_twisted_array"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    found, base_object = utils.find_object(base_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: base_object not in document."))
        return None

    found, path_object = utils.find_object(path_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: path_object not in document."))
        return None
    try:
        utils.type_check([(count, (int, float))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number."))
        return None
    count = int(count)

    use_link = bool(use_link)
    if use_link:
        # The PathTwistedArray class must be called in this special way
        # to make it a PathTwistLinkArray
        new_obj = doc.addObject(
            "Part::FeaturePython", "PathTwistedArray", PathTwistedArray(None), None, True
        )
    else:
        new_obj = doc.addObject("Part::FeaturePython", "PathTwistedArray")
        PathTwistedArray(new_obj)

    new_obj.Base = base_object
    new_obj.PathObject = path_object
    new_obj.Count = count
    new_obj.RotationFactor = rot_factor

    if App.GuiUp:
        if use_link:
            ViewProviderDraftLink(new_obj.ViewObject)
        else:
            ViewProviderDraftArray(new_obj.ViewObject)
            gui_utils.formatObject(new_obj, new_obj.Base)
            new_obj.ViewObject.Proxy.resetColors(new_obj.ViewObject)
        new_obj.Base.ViewObject.hide()
        gui_utils.select(new_obj)

    return new_obj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import os
from builtins import open as pyopen

import FreeCAD

translate = FreeCAD.Qt.translate


def open(filename):
    """opens a SHP/SHX/DBF file in a new FreeCAD document"""

    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    doc = insert(filename, doc.Name)
    return doc


def insert(filename, docname, record=None):
    """imports a SHP/SHX/DBF file in an existing FreeCAD document.
    the record attribute is an optional string indicating the shapefile
    field to use to give elevations to the different shapes. If not used,
    if running in GUI mode, a dialog will pop up to ask the user which
    field to use."""

    if not checkShapeFileLibrary():
        return

    import shapefile
    import Part

    # read the shape file
    # doc at https://github.com/GeospatialPython/pyshp

    shp = shapefile.Reader(filename)

    # check which record to use for elevation
    if not record:
        fields = ["None"] + [field[0] for field in shp.fields]
        if FreeCAD.GuiUp:
            import FreeCADGui
            from PySide import QtGui

            reply = QtGui.QInputDialog.getItem(
                FreeCADGui.getMainWindow(),
                translate("Arch", "Shapes elevation"),
                translate("Arch", "Choose which field provides shapes elevations:"),
                fields,
            )
            if reply[1] and reply[0] != "None":
                record = reply[0]

    # build shapes
    shapes = []
    for shaperec in shp.shapeRecords():
        shape = None
        pts = []
        for p in shaperec.shape.points:
            if len(p) > 2:
                pts.append(FreeCAD.Vector(p[0], p[1], p[2]))
            else:
                pts.append(FreeCAD.Vector(p[0], p[1], 0))
        if shp.shapeTypeName in ["POLYGON", "POLYGONZ"]:
            # faces
            pts.append(pts[0])
            shape = Part.makePolygon(pts)
            shape = Part.Face(shape)
        elif shp.shapeTypeName in ["POINT", "POINTZ"]:
            # points
            verts = [Part.Vertex(p) for p in pts]
            if verts:
                shape = Part.makeCompound(verts)
        else:
            # polylines
            shape = Part.makePolygon(pts)
        if record:
            elev = shaperec.record[record]
            if elev:
                shape.translate(FreeCAD.Vector(0, 0, elev))
        if shape:
            shapes.append(shape)
    if shapes:
        result = Part.makeCompound(shapes)
        obj = FreeCAD.ActiveDocument.addObject("Part::Feature", "shapefile")
        obj.Shape = result
        obj.Label = os.path.splitext(os.path.basename(filename))[0]
        FreeCAD.ActiveDocument.recompute()
    else:
        FreeCAD.Console.PrintWarning(translate("Arch", "No shape found in this file") + "\n")


def getFields(filename):
    """returns the fields found in the given file"""

    if not checkShapeFileLibrary():
        return
    import shapefile

    shp = shapefile.Reader(filename)
    return [field[0] for field in shp.fields]


def checkShapeFileLibrary():
    """Looks for and/or installs the ShapeFile library"""

    try:
        import shapefile
    except Exception:
        # pinning to pyshp upstream v2.4.1 since in 3.0.0 the path to shapefile.py changes
        url = (
            "https://raw.githubusercontent.com/GeospatialPython/pyshp/refs/tags/2.4.1/shapefile.py"
        )
        if FreeCAD.GuiUp:
            import urllib.request
            import FreeCADGui
            from PySide import QtGui

            reply = QtGui.QMessageBox.question(
                FreeCADGui.getMainWindow(),
                translate("Arch", "Shapefile module not found"),
                translate(
                    "Arch",
                    "The shapefile Python library was not found on your system. Would you like to download it now from %1? It will be placed in your macros folder.",
                ).replace(
                    "%1",
                    '<a href="https://github.com/GeospatialPython/pyshp">https://github.com/GeospatialPython/pyshp</a>',
                ),
                QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                QtGui.QMessageBox.No,
            )
            if reply == QtGui.QMessageBox.Yes:
                u = urllib.request.urlopen(url)
                if not u:
                    FreeCAD.Console.PrintError(
                        translate("Arch", "Error: Unable to download from %1").replace("%1", url)
                        + "\n"
                    )
                    return False
                b = u.read()
                fp = os.path.join(FreeCAD.getUserMacroDir(True), "shapefile.py")
                f = pyopen(fp, "wb")
                f.write(b)
                f.close()
            else:
                FreeCAD.Console.PrintError(
                    translate("Arch", "Shapefile module not downloaded. Aborting.") + "\n"
                )
                return False
        else:
            FreeCAD.Console.PrintError(
                translate("Arch", "Shapefile module not found. Aborting.") + "\n"
            )
            FreeCAD.Console.PrintMessage(
                translate(
                    "Arch",
                    "The shapefile library can be downloaded from the following URL and installed in your macros folder:",
                )
                + "\n"
            )
            FreeCAD.Console.PrintMessage(url)
            return False
    return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions for inversive operations using a circle.

http://en.wikipedia.org/wiki/Inversive_geometry
"""
## @package circle_inversion
# \ingroup draftgeoutils
# \brief Provides various functions for inversive geometry operations.

import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils

from draftgeoutils.general import NORM, geomType
from draftgeoutils.geometry import findDistance

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def pointInversion(circle, point):
    """Return the circle inversion of a point.

    It will return `None` if the given point is equal to the center
    of the circle.
    """
    if geomType(circle) != "Circle" or isinstance(point, App.Vector):
        print("debug: pointInversion bad parameters!")
        return None

    cen = circle.Curve.Center
    rad = circle.Curve.Radius

    if DraftVecUtils.equals(cen, point):
        return None

    # Inverse the distance of the point
    # dist(cen -> P) = r^2 / dist(cen -> invP)

    dist = DraftVecUtils.dist(point, cen)
    invDist = rad**2 / dist

    invPoint = App.Vector(0, 0, point.z)
    invPoint.x = cen.x + (point.x - cen.x) * invDist / dist
    invPoint.y = cen.y + (point.y - cen.y) * invDist / dist

    return invPoint


def polarInversion(circle, edge):
    """Return the inversion pole of a line. The edge is the polar.

    The nearest point on the line is inversed.

    http://mathworld.wolfram.com/InversionPole.html
    """
    if geomType(circle) != "Circle" or geomType(edge) != "Line":
        print("debug: circleInversionPole bad parameters! Must be a circle.")
        return None

    nearest = circle.Curve.Center.add(findDistance(circle.Curve.Center, edge, False))
    if nearest:
        inversionPole = pointInversion(circle, nearest)
        if inversionPole:
            return inversionPole

    return None


def circleInversion(circle, circle2):
    """Circle inversion of a circle, inverting the center point.

    Returns the new circle created from the inverted center of circle2.
    """
    if geomType(circle) != "Circle" or geomType(circle2) != "Circle":
        print("debug: circleInversion bad parameters! Must be circles.")
        return None

    cen1 = circle.Curve.Center

    cen2 = circle2.Curve.Center
    rad2 = circle2.Curve.Radius

    if DraftVecUtils.equals(cen1, cen2):
        return None

    invCen2 = pointInversion(circle, cen2)

    pointOnCircle2 = App.Vector.add(cen2, App.Vector(rad2, 0, 0))
    invPointOnCircle2 = pointInversion(circle, pointOnCircle2)

    return Part.Circle(invCen2, NORM, DraftVecUtils.dist(invCen2, invPointOnCircle2))


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *   Copyright (c) 2024 FreeCAD Project Association                        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Sketch objects from Draft objects."""
## @package make_sketch
# \ingroup draftmake
# \brief Provides functions to create Sketch objects from Draft objects.

## \addtogroup draftmake
# @{
import math

import FreeCAD as App
import DraftVecUtils
import DraftGeomUtils
from draftutils import gui_utils
from draftutils import utils
from draftutils.translate import translate


def make_sketch(
    objects_list,
    autoconstraints=False,
    addTo=None,
    delete=False,
    name="Sketch",
    radiusPrecision=-1,
    tol=1e-3,
):
    """make_sketch(objects_list, [autoconstraints], [addTo], [delete],
                   [name], [radiusPrecision], [tol])

    Makes a Sketch objects_list with the given Draft objects.

    Parameters
    ----------
    objects_list: can be single or list of objects of Draft type objects,
        Part::Feature, Part.Shape, or mix of them.

    autoconstraints(False): if True, coincident, horizontal and vertical
        constraints will be added automatically.

    addTo(None): if set to an existing sketch, geometry will be added to it
        instead of creating a new one.

    delete(False): if True, the original object will be deleted.
        If set to a string "all" the object and all its linked object will be
        deleted.

    name("Sketch"): the name for the new sketch object.

    radiusPrecision(-1): If <0, disable radius constraint. If =0, add individual
        radius constraint. If >0, the radius will be rounded according to this
        precision, and "Equal" constraint will be added to curve with equal
        radius within precision.

    tol(1e-3): Tolerance used to check if the shapes are planar and coplanar.
        Consider change to tol=-1 for a more accurate analysis.
    """

    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return

    import Part
    from Sketcher import Constraint

    # lists to accumulate shapes with defined normal and undefined normal
    shape_norm_yes = list()
    shape_norm_no = list()

    if not isinstance(objects_list, (list, tuple)):
        objects_list = [objects_list]

    for obj in objects_list:
        if isinstance(obj, Part.Shape):
            shape = obj
        elif not hasattr(obj, "Shape"):
            App.Console.PrintError(translate("draft", "No shape found") + "\n")
            return None
        else:
            shape = obj.Shape

        if not DraftGeomUtils.is_planar(shape, tol):
            App.Console.PrintError(translate("draft", "All shapes must be planar") + "\n")
            return None

        if DraftGeomUtils.get_normal(shape, tol):
            shape_norm_yes.append(shape)
        else:
            shape_norm_no.append(shape)

    shapes_list = shape_norm_yes + shape_norm_no

    # test if all shapes are coplanar
    if len(shape_norm_yes) >= 1:
        for shape in shapes_list[1:]:
            if not DraftGeomUtils.are_coplanar(shapes_list[0], shape, tol):
                App.Console.PrintError(translate("draft", "All shapes must be coplanar") + "\n")
                return None
        # define sketch normal
        normal = DraftGeomUtils.get_normal(shapes_list[0], tol)

    else:
        # suppose all geometries are straight lines or points
        points = [vertex.Point for shape in shapes_list for vertex in shape.Vertexes]
        if len(points) >= 2:
            try:
                poly = Part.makePolygon(points)
            except Part.OCCError:
                # all points coincide
                normal = App.Vector(0, 0, 1)
            else:
                if not DraftGeomUtils.is_planar(poly, tol):
                    App.Console.PrintError(translate("draft", "All shapes must be coplanar") + "\n")
                    return None
                normal = DraftGeomUtils.get_shape_normal(poly)
        else:
            # only one point
            normal = App.Vector(0, 0, 1)

    if addTo:
        nobj = addTo
    else:
        nobj = App.ActiveDocument.addObject("Sketcher::SketchObject", name)

    # Collect constraints and add in one go to improve performance
    constraints = []
    radiuses = {}

    def addRadiusConstraint(edge):
        try:
            if radiusPrecision < 0:
                return
            if radiusPrecision == 0:
                constraints.append(Constraint("Radius", nobj.GeometryCount - 1, edge.Curve.Radius))
                return
            r = round(edge.Curve.Radius, radiusPrecision)
            constraints.append(Constraint("Equal", radiuses[r], nobj.GeometryCount - 1))
        except KeyError:
            radiuses[r] = nobj.GeometryCount - 1
            constraints.append(Constraint("Radius", nobj.GeometryCount - 1, r))
        except AttributeError:
            pass

    def convertBezier(edge):
        if DraftGeomUtils.geomType(edge) == "BezierCurve":
            return edge.Curve.toBSpline(edge.FirstParameter, edge.LastParameter).toShape()
        else:
            return edge

    axis = App.Vector(0, 0, 1).cross(normal)
    if axis.Length > 1e-6:
        axis.normalize()
    elif normal.z >= 0:
        axis = App.Vector(0, 0, 1)
    else:
        axis = App.Vector(0, 0, -1)
    angle = math.degrees(DraftVecUtils.angle(normal, App.Vector(0, 0, 1)))
    rotation = App.Rotation(axis, angle)

    point = shapes_list[0].Vertexes[0].Point
    base = App.Vector(normal)
    base.Length = point.dot(
        base.normalize()
    )  # See https://forum.freecad.org/viewtopic.php?f=22&t=69304#p601149

    nobj.Placement = App.Placement(base, rotation)

    for obj in objects_list:
        ok = False
        tp = utils.get_type(obj)

        if tp == "Point":
            # obj.Shape.copy() does not work properly for a Draft_Point.
            # The coords of the point are multiplied by 2.
            # We therefore create a Part Vertex instead.
            shape = Part.Vertex(obj.Shape.Point)
            if angle:
                shape.rotate(App.Vector(0, 0, 0), axis, -angle)
            point = Part.Point(shape.Point)
            nobj.addGeometry(point)
            ok = True

        elif tp == "Shape" or hasattr(obj, "Shape"):
            shape = obj if tp == "Shape" else obj.Shape
            for e in shape.Edges:
                newedge = convertBezier(e)
                nobj.addGeometry(DraftGeomUtils.orientEdge(newedge, normal, make_arc=True))
                addRadiusConstraint(newedge)
            ok = True

        if ok and delete:

            def delObj(obj):
                if obj.InList:
                    App.Console.PrintWarning(
                        translate(
                            "draft", "Cannot delete object {} with dependency".format(obj.Label)
                        )
                        + "\n"
                    )
                else:
                    obj.Document.removeObject(obj.Name)

            try:
                if delete == "all":
                    objs = [obj]
                    while objs:
                        obj = objs[0]
                        objs = objs[1:] + obj.OutList
                        delObj(obj)
                else:
                    delObj(obj)
            except Exception as ex:
                App.Console.PrintWarning(
                    translate("draft", "Failed to delete object {}: {}".format(obj.Label, ex))
                    + "\n"
                )

    nobj.addConstraint(constraints)
    if autoconstraints:
        nobj.detectMissingPointOnPointConstraints(utils.tolerance())
        nobj.makeMissingPointOnPointCoincident(False)
        nobj.detectMissingVerticalHorizontalConstraints(utils.tolerance())
        nobj.makeMissingVerticalHorizontal(False)
        # The MissingVerticalHorizontal functions do not work properly.
        # If elements are added to an existing sketch redundant constraints are created.
        # This can happen if DXF files are imported with the legacy importer.
        # https://forum.freecad.org/viewtopic.php?t=97072
        # https://github.com/FreeCAD/FreeCAD/issues/21396
        # To address this we check for redundant constraints. This can be necessary even
        # the functions were to work properly. For example with this scenario:
        # https://github.com/FreeCAD/FreeCAD/issues/19978
        nobj.solve()
        for idx in nobj.RedundantConstraints[::-1]:
            # Output of RedundantConstraints is one-based.
            nobj.delConstraint(idx - 1)

    return nobj


makeSketch = make_sketch

## @}



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Query FEM specific settings including solver settings.

Query settings from the hierarchically organized settings/parameter system of
FreeCAD related to the FEM module. The collection of independent functions use
the settings system as a backend and expose a easy to use interface for other
modules of the FEM module.

Functions querying solver specific settings always take a solver name as a
string to identify the solver in question. At the moment the following solvers
are supported:

    - Calculix
    - ElmerSolver
    - Mystran
    - Z88

To query settings about those solver the solver name must be given exactly in
the form written in the list above. To make the solver recognize settings for a
new solver have a look at :class:`_SolverDlg`.
"""

__title__ = "FreeCAD FEM solver settings"
__author__ = "Markus Hovorka, Bernd Hahnebach"
__url__ = "https://www.freecad.org"


import FreeCAD


class DirSetting:
    """Enum of possible directory setting values.

    Strings used to indicate the solver directory setting set in FreeCADs
    setting system. Returned by :func:`get_dir_setting` for that purpose. There
    are three different possible values:

    :cvar TEMPORARY:
        Let FreeCAD manage (create, delete) the working directories for all
        solver. Use temporary directories.

    :cvar BESIDE:
        Create a directory in the same folder in which the FCStd file of the
        document is located. Use Subfolder for each solver (e.g. for a file
        ./mydoc.FCStd and a solver with the label Elmer002 use
        ./mydoc/Elmer002).

    :cvar CUSTOM:
        Use directory set below. Create own subdirectory for every solver. Name
        directory after the solver label prefixed with the document name.
    """

    TEMPORARY = "temporary"
    BESIDE = "beside"
    CUSTOM = "custom"


# FEM parameter location path
_PARAM_PATH = "User parameter:BaseApp/Preferences/Mod/Fem/"
_GENERAL_PARAM = _PARAM_PATH + "General"


def get_binary(name, silent=False):
    """Find binary of solver *name* honoring user settings.

    Return the specific path set by the user in FreeCADs settings/parameter
    system if set or the default binary name if no specific path is set. If no
    path was found because the solver *name* is not supported ``None`` is
    returned.
    This method does not check whether the binary actually exists and is callable.
    That check is done in DlgSettingsFem_Solver_Imp.cpp

    :param name: solver id as a ``str`` (see :mod:`femsolver.settings`)
    :param silent: whether to output error if binary not found
    """
    if name in _SOLVER_PARAM:
        binary = _SOLVER_PARAM[name].get_binary(silent)
        return binary
    else:
        if not silent:
            FreeCAD.Console.PrintError(
                "Settings solver name: {} not found in "
                "solver settings modules _SOLVER_PARAM dirctionary.\n".format(name)
            )
        return None


def get_cores(name):
    """Read number of CPU cores for solver *name* honoring user settings.

    Returns number of CPU cores to be used for the solver run

    :param name: solver id as a ``str`` (see :mod:`femsolver.settings`)
    """
    if name in _SOLVER_PARAM:
        cores = _SOLVER_PARAM[name].get_cores()
        return cores
    else:
        FreeCAD.Console.PrintError(
            "Settings solver name: {} not found in "
            "solver settings modules _SOLVER_PARAM dirctionary.\n".format(name)
        )
        return None


def get_write_comments(name):
    """Check whether "write_comments" is set for solver.

    Returns ``True`` if the "write_comments" setting/parameter is set for the
    solver with the id *name*. Returns ``False`` otherwise. If the solver is
    not supported ``None`` is returned.

    :param name: solver id as a ``str`` (see :mod:`femsolver.settings`)
    """
    if name in _SOLVER_PARAM:
        return _SOLVER_PARAM[name].get_write_comments()
    else:
        FreeCAD.Console.PrintError(
            "Settings solver name: {} not found in "
            "solver settings modules _SOLVER_PARAM dirctionary.\n".format(name)
        )
        return None


def get_custom_dir():
    """Get value for :term:`General/CustomDirectoryPath` parameter."""
    param_group = FreeCAD.ParamGet(_GENERAL_PARAM)
    return param_group.GetString("CustomDirectoryPath")


def get_dir_setting():
    """Return directory setting set by the user.

    Return one of the three possible values of the :class:`DirSetting` enum
    depending on the setting set in FreeCAD parameter system. Result dependes
    on the values of :term:`General/UseTempDirectory`,
    :term:`General/UseBesideDirectory` and :term:`General/UseCustomDirectory`.
    """
    param_group = FreeCAD.ParamGet(_GENERAL_PARAM)
    if param_group.GetBool("UseBesideDirectory"):
        return DirSetting.BESIDE
    elif param_group.GetBool("UseCustomDirectory"):
        return DirSetting.CUSTOM
    return DirSetting.TEMPORARY


def get_default_solver():
    """Return default solver name."""
    solver_map = {0: None}
    if get_binary("Calculix", True):
        solver_map[1] = "CalculiX"
    if get_binary("ElmerSolver", True):
        solver_map[len(solver_map)] = "Elmer"
    if get_binary("Mystran", True):
        solver_map[len(solver_map)] = "Mystran"
    if get_binary("Z88", True):
        solver_map[len(solver_map)] = "Z88"
    param_group = FreeCAD.ParamGet(_GENERAL_PARAM)
    return solver_map[param_group.GetInt("DefaultSolver", 0)]


class _SolverDlg:
    """Internal query logic for solver specific settings.

    Each instance queries settings for one specific solver (e.g. Elmer) common
    among all solvers. To clarify: There are a few settings that are useful
    for every solver (e.g. where to find the solver binary) but the value and
    the FreeCAD parameter path is different for each one. A instance of this
    class contains all the solver specific paths needed. The settings can be
    queried via the methods which use those path members to query the value for
    the specific solver.

    :ivar default:
        Default binary name as a string preferably without a prefix path to
        make it more generic (e.g. "ccx"). This only works if the binary can be
        found via the PATH environment variable on linux or similar mechanisms
        on other operating systems. Used if nothing else is specified by the
        user.

    :ivar param_path:
        Parent param path (FreeCADs settings/parameter system) that contains
        all settings for the specific solver.

    :ivar use_default:
        Param path identifying the "use_default" setting. Only specifie the
        last part as the *param_path* is prepended to this value.

    :ivar custom_path:
        Param path identifying the "custom_path" setting. Only specifie the
        last part as the *param_path* is prepended to this value.
    """

    WRITE_COMMENTS_PARAM = "writeCommentsToInputFile"

    def __init__(self, default, param_path, custom_path):
        self.default = default
        self.param_path = param_path
        self.custom_path = custom_path

        self.param_group = FreeCAD.ParamGet(self.param_path)

    def get_binary(self, silent=False):

        # set the binary path to the FreeCAD defaults
        # ATM pure unix shell commands without path names are used as standard
        # TODO the binaries provided with the FreeCAD distribution should be found
        # without any additional user input
        # see ccxttols, it works for Windows and Linux there
        binary = self.param_group.GetString(self.custom_path)
        if not binary:
            binary = self.default
        FreeCAD.Console.PrintLog(f"Solver binary path default: {binary} \n")

        # get the whole binary path name for the given command or binary path and return it
        # None is returned if the binary has not been found
        # The user does not know what exactly has going wrong.
        from shutil import which as find_bin

        the_found_binary = find_bin(binary)
        if (the_found_binary is None) and (not silent):
            FreeCAD.Console.PrintError(
                f"The binary has not been found. Full binary search path: {binary}\n"
            )
        else:
            FreeCAD.Console.PrintLog(f"Found solver binary path: {the_found_binary}\n")
        return the_found_binary

    def get_cores(self):
        cores = self.param_group.GetInt("UseNumberOfCores")
        return cores

    def get_write_comments(self):
        return self.param_group.GetBool(self.WRITE_COMMENTS_PARAM, True)


_SOLVER_PARAM = {
    "Calculix": _SolverDlg(
        default="ccx",
        param_path=_PARAM_PATH + "Ccx",
        custom_path="ccxBinaryPath",
    ),
    "ElmerSolver": _SolverDlg(
        default="ElmerSolver",
        param_path=_PARAM_PATH + "Elmer",
        custom_path="elmerBinaryPath",
    ),
    "ElmerGrid": _SolverDlg(
        default="ElmerGrid",
        param_path=_PARAM_PATH + "Elmer",
        custom_path="gridBinaryPath",
    ),
    "Mystran": _SolverDlg(
        default="mystran",
        param_path=_PARAM_PATH + "Mystran",
        custom_path="mystranBinaryPath",
    ),
    "Z88": _SolverDlg(
        default="z88r",
        param_path=_PARAM_PATH + "Z88",
        custom_path="z88BinaryPath",
    ),
}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM DimensionAligned command"""

import FreeCAD
import FreeCADGui

from draftguitools import gui_dimensions

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_DimensionAligned(gui_dimensions.Dimension):

    def __init__(self):
        super().__init__()

    def GetResources(self):
        return {
            "Pixmap": "BIM_DimensionAligned",
            "MenuText": QT_TRANSLATE_NOOP("BIM_DimensionAligned", "Aligned Dimension"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_DimensionAligned", "Creates an aligned dimension"),
            "Accel": "D, I",
        }


class BIM_DimensionHorizontal(gui_dimensions.Dimension):

    def __init__(self):
        super().__init__()

    def GetResources(self):
        return {
            "Pixmap": "BIM_DimensionHorizontal.svg",
            "MenuText": QT_TRANSLATE_NOOP("BIM_DimensionHorizontal", "Horizontal Dimension"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_DimensionHorizontal", "Creates an horizontal dimension"
            ),
            "Accel": "D, H",
        }

    def Activated(self):
        import WorkingPlane

        self.dir = WorkingPlane.get_working_plane().u
        super().Activated(dir_vec=self.dir)


class BIM_DimensionVertical(gui_dimensions.Dimension):

    def __init__(self):
        super().__init__()

    def GetResources(self):
        return {
            "Pixmap": "BIM_DimensionVertical",
            "MenuText": QT_TRANSLATE_NOOP("BIM_DimensionVertical", "Vertical Dimension"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_DimensionVertical", "Creates a vertical dimension"),
            "Accel": "D, V",
        }

    def Activated(self):
        import WorkingPlane

        self.dir = WorkingPlane.get_working_plane().v
        super().Activated(dir_vec=self.dir)


FreeCADGui.addCommand("BIM_DimensionVertical", BIM_DimensionVertical())
FreeCADGui.addCommand("BIM_DimensionHorizontal", BIM_DimensionHorizontal())
FreeCADGui.addCommand("BIM_DimensionAligned", BIM_DimensionAligned())



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2024 The FreeCAD Project Association AISBL               *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import FreeCAD


def _remove_from_list(prefs, pref_name):
    # Remove Start and Web from a preference that consists of a comma-separated list of workbenches
    # and ensure that the list was not empty

    is_empty_string = "<IS_EMPTY>"
    mods = prefs.GetString(pref_name, is_empty_string).split(",")

    if is_empty_string in mods:
        return

    if "StartWorkbench" in mods:
        mods.remove("StartWorkbench")
    if "WebWorkbench" in mods:
        mods.remove("WebWorkbench")

    prefs.SetString(pref_name, ",".join(mods))


class StartMigrator2024:
    """In April 2024 the old Start workbench was retired, and replaced with the current Start command. This function
    cleans up references to the old workbench, migrating the settings where appropriate, and deleting them where they
    are no longer relevant. Web was also removed, without replacement.
    TODO: Remove the 2024 migration code when enough time has passed that it is no longer useful"""

    def __init__(self):
        self.completion_indicator = "Migration2024Complete"
        self.start_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start")
        self.general_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/General")
        self.migration_2024_complete = self.start_prefs.GetBool(self.completion_indicator, False)

    def run_migration(self):
        if self.migration_2024_complete:
            # Refuse to run if it's already been done
            return
        FreeCAD.Console.PrintMessage("Migrating Start Workbench to Start command... ")
        self._update_startup_flags()
        self._remove_wb_references()
        self._remove_commands()
        self._remove_toolbars()
        self._remove_deprecated_parameters()
        self._mark_complete()
        FreeCAD.Console.PrintMessage("done.\n")

    # If the old Start workbench was set as the Autoload Module, reconfigure it so the Start command is run at startup,
    # and set the Autoload module to "PartDesignWorkbench"
    def _update_startup_flags(self):
        autoload_module = self.general_prefs.GetString("AutoloadModule", "StartWorkbench")
        if autoload_module == "StartWorkbench":
            self.start_prefs.SetBool("ShowOnStartup", True)
            self.general_prefs.SetString("AutoloadModule", "PartDesignWorkbench")
        else:
            self.start_prefs.SetBool("ShowOnStartup", False)

    # Remove all references to the StartWorkbench in the various workbench lists
    def _remove_wb_references(self):
        general_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/General")
        last_module = general_prefs.GetString("LastModule", "")
        if last_module in ["StartWorkbench", "WebWorkbench"]:
            general_prefs.RemString("LastModule")
        _remove_from_list(general_prefs, "BackgroundAutoloadModules")
        treeview_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Workbenches")
        _remove_from_list(treeview_prefs, "Ordered")
        _remove_from_list(treeview_prefs, "Disabled")

    # Remove any commands that begin with "Start_" or "Web_" (except "Start_Start", the current command)
    def _remove_commands(self):
        command_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Commands")
        commands = command_prefs.GetStrings()
        for command in commands:
            if command.startswith("Web_") or (
                command.startswith("Start_") and command != "Start_Start"
            ):
                command_prefs.RemString(command)

    # Remove any persistent toolbars
    def _remove_toolbars(self):
        tux_prefs = FreeCAD.ParamGet("User parameter:Tux/PersistentToolbars/User")
        groups = tux_prefs.GetGroups()
        if "StartWorkbench" in groups:
            tux_prefs.RemGroup("StartWorkbench")
        if "WebWorkbench" in groups:
            tux_prefs.RemGroup("WebWorkbench")

    # Delete old Start preferences
    def _remove_deprecated_parameters(self):
        show_on_startup = self.start_prefs.GetBool("ShowOnStartup", True)
        show_examples = self.start_prefs.GetBool("ShowExamples", True)
        close_start = self.start_prefs.GetBool("closeStart", False)
        custom_folder = self.start_prefs.GetString("ShowCustomFolder", "")
        self.start_prefs.Clear()
        self.start_prefs.SetBool("ShowOnStartup", show_on_startup)
        self.start_prefs.SetBool("ShowExamples", show_examples)
        self.start_prefs.SetBool("CloseStart", close_start)
        # ShowCustomFolder renamed to CustomFolder in 1.1
        self.start_prefs.SetString("CustomFolder", custom_folder)

    # Indicate that this migration has been run
    def _mark_complete(self):
        self.start_prefs.SetBool(self.completion_indicator, True)



# SPDX-License-Identifier: LGPL-2.1-or-later

# *****************************************************************************
# *                                                                           *
# *   (c) sliptonic (shopinthewoods@gmail.com) 2014                           *
# *   (c) Gauthier Briere - 2018, 2019                                        *
# *   (c) Schildkroet - 2019-2020                                             *
# *   (c) Gary L Hasson - 2020                                                *
# *                                                                           *
# *   This file is part of the FreeCAD CAx development system.                *
# *                                                                           *
# *   This program is free software; you can redistribute it and/or modify    *
# *   it under the terms of the GNU Lesser General Public License (LGPL)      *
# *   as published by the Free Software Foundation; either version 2 of       *
# *   the License, or (at your option) any later version.                     *
# *   for detail see the LICENCE text file.                                   *
# *                                                                           *
# *   FreeCAD is distributed in the hope that it will be useful,              *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
# *   GNU Lesser General Public License for more details.                     *
# *                                                                           *
# *   You should have received a copy of the GNU Library General Public       *
# *   License along with FreeCAD; if not, write to the Free Software          *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    *
# *   USA                                                                     *
# *                                                                           *
# *****************************************************************************


from datetime import datetime
import argparse
import shlex
import FreeCAD
from FreeCAD import Units
import Path
import Path.Base.Util as PathUtil
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
from builtins import open as pyopen

Revised = "2020-11-03"  # Revision date for this file.

# *****************************************************************************
# *   Due to the fundamentals of the FreeCAD pre-processor,                   *
# *   this post processor can only operate in the following modes:            *
# *   G90 Absolute positions                                                  *
# *   G21 Metric units (mm)                                                   *
# *   G17 XY plane (3 axis vertical milling only)                             *
# *                                                                           *
# *****************************************************************************


TOOLTIP = """
Generate g-code from a Path that is compatible with the Marlin controller.
import marlin_post
marlin_post.export(object, "/path/to/file.nc")
"""

# *****************************************************************************
# * Initial configuration, not changeable                                     *
# *****************************************************************************
MOTION_MODE = "G90"  # G90 only, for absolute moves
WORK_PLANE = "G17"  # G17 only, XY plane, for vertical milling
UNITS = "G21"  # G21 only, for metric
UNIT_FORMAT = "mm"
UNIT_FEED_FORMAT = "mm/min"

# *****************************************************************************
# * Initial configuration, changeable via command line arguments              *
# *****************************************************************************
PRECISION = 3  # Decimal places displayed for metric
DRILL_RETRACT_MODE = "G98"  # End of drill-cycle retractation type. G99
# is the alternative.
TRANSLATE_DRILL_CYCLES = True  # If true, G81, G82, and G83 are translated
# into G0/G1 moves
OUTPUT_TOOL_CHANGE = False  # Do not output M6 tool change (comment it)
RETURN_TO = None  # None = No movement at end of program
SPINDLE_WAIT = 3  # 0 == No waiting after M3 / M4
MODAL = False  # True: Commands are suppressed if they are
# the same as the previous line
LINENR = 100  # Line number starting value
LINEINCR = 10  # Line number increment
PRE_OPERATION = """"""  # Pre operation text will be inserted before
# every operation
POST_OPERATION = """"""  # Post operation text will be inserted after
# every operation
TOOL_CHANGE = """"""  # Tool Change commands will be inserted
# before a tool change

# *****************************************************************************
# * Initial gcode output options, changeable via command line arguments       *
# *****************************************************************************
OUTPUT_HEADER = True  # Output header in output gcode file
OUTPUT_COMMENTS = True  # Comments in output gcode file
OUTPUT_FINISH = False  # Include an operation finished comment
OUTPUT_PATH = False  # Include a Path: comment
OUTPUT_MARLIN_CONFIG = False  # Display expected #defines for Marlin config
OUTPUT_LINE_NUMBERS = False  # Output line numbers in output gcode file
OUTPUT_BCNC = False  # Add bCNC operation block headers in output
# gcode file
SHOW_EDITOR = True  # Display the resulting gcode file

# *****************************************************************************
# * Command line arguments                                                    *
# *****************************************************************************
parser = argparse.ArgumentParser(prog="marlin", add_help=False)
parser.add_argument("--header", action="store_true", help="output headers (default)")
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--comments", action="store_true", help="output comment (default)")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument("--finish-comments", action="store_true", help="output finish-comment")
parser.add_argument(
    "--no-finish-comments",
    action="store_true",
    help="suppress finish-comment output (default)",
)
parser.add_argument("--path-comments", action="store_true", help="output path-comment")
parser.add_argument(
    "--no-path-comments",
    action="store_true",
    help="suppress path-comment output (default)",
)
parser.add_argument("--marlin-config", action="store_true", help="output #defines for Marlin")
parser.add_argument(
    "--no-marlin-config",
    action="store_true",
    help="suppress output #defines for Marlin (default)",
)
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-line-numbers",
    action="store_true",
    help="do not prefix with line numbers (default)",
)
parser.add_argument(
    "--show-editor",
    action="store_true",
    help="pop up editor before writing output (default)",
)
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="do not pop up editor before writing output",
)
parser.add_argument("--precision", default="3", help="number of digits of precision, default=3")
parser.add_argument(
    "--translate_drill",
    action="store_true",
    help="translate drill cycles G81, G82, G83 into G0/G1 movements (default)",
)
parser.add_argument(
    "--no-translate_drill",
    action="store_true",
    help="do not translate drill cycles G81, G82, G83 into G0/G1 movements",
)
parser.add_argument(
    "--preamble", help='set commands to be issued before the first command, default=""'
)
parser.add_argument(
    "--postamble", help='set commands to be issued after the last command, default="M5\\n"'
)
parser.add_argument("--tool-change", action="store_true", help="Insert M6 for all tool changes")
parser.add_argument(
    "--wait-for-spindle",
    type=int,
    default=3,
    help="Wait for spindle to reach desired speed after M3 or M4, default=0",
)
parser.add_argument(
    "--return-to",
    default="",
    help='When done, move to, e.g. --return-to="3.175, 4.702, 50.915"',
)
parser.add_argument(
    "--bcnc",
    action="store_true",
    help="Add Job operations as bCNC block headers. \
    Consider suppressing existing comments: Add argument --no-comments",
)
parser.add_argument(
    "--no-bcnc", action="store_true", help="suppress bCNC block header output (default)"
)
TOOLTIP_ARGS = parser.format_help()

# *****************************************************************************
# * Marlin 2.x:                                                               *
# * Ignores commands that it does not implement.                              *
# * Some machining-related commands may conflict with gcodes that Marlin      *
# * has assigned to 3D printing commands.                                     *
# * Therefore, check FreeCAD gcodes for conflicts with Marlin.                *
# * Marlin 2.x ignores the ENTIRE COMMAND LINE if there is more than          *
# * one command per line.                                                     *
# *****************************************************************************

# Default preamble text will appear at the beginning of the gcode output file.
PREAMBLE = """"""

# Default postamble text will appear following the last operation.
POSTAMBLE = """M5
"""

# *****************************************************************************
# * Internal global variables                                                 *
# *****************************************************************************
MOTION_COMMANDS = ["G0", "G00", "G1", "G01", "G2", "G02", "G3", "G03"]
RAPID_MOVES = ["G0", "G00"]  # Rapid moves gcode commands definition
SUPPRESS_COMMANDS = [""]  # These commands are ignored by commenting them out
COMMAND_SPACE = " "
# Global variables storing current position (Use None for safety.)
CURRENT_X = None
CURRENT_Y = None
CURRENT_Z = None


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_FINISH
    global OUTPUT_PATH
    global OUTPUT_MARLIN_CONFIG
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_FEED_FORMAT
    global UNIT_FORMAT
    global TRANSLATE_DRILL_CYCLES
    global OUTPUT_TOOL_CHANGE
    global SPINDLE_WAIT
    global RETURN_TO
    global OUTPUT_BCNC
    global PRECISION

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.header:
            OUTPUT_HEADER = True
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.comments:
            OUTPUT_COMMENTS = True
        if args.no_finish_comments:
            OUTPUT_FINISH = False
        if args.finish_comments:
            OUTPUT_FINISH = True
        if args.no_path_comments:
            OUTPUT_PATH = False
        if args.path_comments:
            OUTPUT_PATH = True
        if args.no_marlin_config:
            OUTPUT_MARLIN_CONFIG = False
        if args.marlin_config:
            OUTPUT_MARLIN_CONFIG = True
        if args.no_line_numbers:
            OUTPUT_LINE_NUMBERS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        if args.show_editor:
            SHOW_EDITOR = True
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        if args.no_translate_drill:
            TRANSLATE_DRILL_CYCLES = False
        if args.translate_drill:
            TRANSLATE_DRILL_CYCLES = True
        if args.tool_change:
            OUTPUT_TOOL_CHANGE = True
        if args.return_to:
            RETURN_TO = args.return_to
            if RETURN_TO.find(",") == -1:
                RETURN_TO = None
                print("--return-to coordinates must be specified as:")
                print('--return-to "x.n,y.n,z.n"')
        if args.bcnc:
            OUTPUT_BCNC = True
        if args.no_bcnc:
            OUTPUT_BCNC = False
        SPINDLE_WAIT = args.wait_for_spindle
        PRECISION = args.precision

    except Exception as e:
        return False

    return True


# For debug...
def dump(obj):
    for attr in dir(obj):
        try:
            if attr.startswith("__"):
                continue
            print(">" + attr + "<")
            attr_text = "%s = %s" % (attr, getattr(obj, attr))
            if attr in ["HorizFeed", "VertFeed"]:
                print("==============\n", attr_text)
                if "mm/s" in attr_text:
                    print("===> metric values <===")
        except Exception:  # Insignificant errors
            # print('==>', obj, attr)
            pass


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None

    global UNITS
    global UNIT_FORMAT
    global UNIT_FEED_FORMAT
    global MOTION_MODE
    global SUPPRESS_COMMANDS

    print("Post Processor: " + __name__ + " postprocessing...")
    gcode = ""

    # Write header:
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__
        gcode += ".py, version: " + Revised + ")\n"
        gcode += linenumber() + "(Output Time:" + str(datetime.now()) + ")\n"

    # Suppress drill-cycle commands:
    if TRANSLATE_DRILL_CYCLES:
        SUPPRESS_COMMANDS += ["G80", "G98", "G99"]

    # Write the preamble:
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(Begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    # Write these settings AFTER the preamble,
    # to prevent the preamble from changing these:
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(Default Configuration)\n"
    gcode += linenumber() + MOTION_MODE + "\n"
    gcode += linenumber() + UNITS + "\n"
    gcode += linenumber() + WORK_PLANE + "\n"

    for obj in objectslist:
        # Debug...
        # print('\n' + '*'*70 + '\n')
        # dump(obj)
        # print('\n' + '*'*70 + '\n')
        if not hasattr(obj, "Path"):
            print(
                "The object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return

        # Skip inactive operations:
        if not PathUtil.activeForOp(obj):
            continue

        # Do the pre_op:
        if OUTPUT_BCNC:
            gcode += linenumber() + "(Block-name: " + obj.Label + ")\n"
            gcode += linenumber() + "(Block-expand: 0)\n"
            gcode += linenumber() + "(Block-enable: 1)\n"
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(Begin operation: " + obj.Label + ")\n"
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # Get coolant mode:
        coolantMode = PathUtil.coolantModeForOp(obj)

        # Turn coolant on if required:
        if OUTPUT_COMMENTS:
            if not coolantMode == "None":
                gcode += linenumber() + "(Coolant On:" + coolantMode + ")\n"
        if coolantMode == "Flood":
            gcode += linenumber() + "M8\n"
        if coolantMode == "Mist":
            gcode += linenumber() + "M7\n"

        # Parse the op:
        gcode += parse(obj)

        # Do the post_op:
        if OUTPUT_COMMENTS and OUTPUT_FINISH:
            gcode += linenumber() + "(Finish operation: " + obj.Label + ")\n"
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # Turn coolant off if previously enabled:
        if not coolantMode == "None":
            if OUTPUT_COMMENTS:
                gcode += linenumber() + "(Coolant Off:" + coolantMode + ")\n"
            gcode += linenumber() + "M9\n"

    # Do the post_amble:
    if OUTPUT_BCNC:
        gcode += linenumber() + "(Block-name: post_amble)\n"
        gcode += linenumber() + "(Block-expand: 0)\n"
        gcode += linenumber() + "(Block-enable: 1)\n"
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(Begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    # Optionally add a final XYZ position to the end of the gcode:
    if RETURN_TO:
        first_comma = RETURN_TO.find(",")
        last_comma = RETURN_TO.rfind(",")  # == first_comma if only one comma
        ref_X = " X" + RETURN_TO[0:first_comma].strip()

        # Z is optional:
        if last_comma != first_comma:
            ref_Z = " Z" + RETURN_TO[last_comma + 1 :].strip()
            ref_Y = " Y" + RETURN_TO[first_comma + 1 : last_comma].strip()
        else:
            ref_Z = ""
            ref_Y = " Y" + RETURN_TO[first_comma + 1 :].strip()

        gcode += linenumber() + "G0" + ref_X + ref_Y + ref_Z + "\n"

    # Optionally add recommended Marlin 2.x configuration to gcode file:
    if OUTPUT_MARLIN_CONFIG:
        gcode += linenumber() + "(Marlin 2.x Configuration)\n"
        gcode += linenumber() + "(The following should be enabled in)\n"
        gcode += linenumber() + "(the configuration files of Marlin 2.x)\n"
        gcode += linenumber() + "(#define ARC_SUPPORT)\n"
        gcode += linenumber() + "(#define CNC_COORDINATE_SYSTEMS)\n"
        gcode += linenumber() + "(#define PAREN_COMMENTS)\n"
        gcode += linenumber() + "(#define GCODE_MOTION_MODES)\n"
        gcode += linenumber() + "(#define G0_FEEDRATE)\n"
        gcode += linenumber() + "(define VARIABLE_G0_FEEDRATE)\n"

    # Show the gcode result dialog:
    if FreeCAD.GuiUp and SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    print("Done postprocessing.")

    # Write the file:
    if not filename == "-":
        gfile = pyopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    if not OUTPUT_LINE_NUMBERS:
        return ""
    global LINENR
    global LINEINCR
    LINENR += LINEINCR
    return "N" + str(LINENR) + " "


def format_outlist(strTable):
    # construct the line for the final output
    global COMMAND_SPACE
    s = ""
    for w in strTable:
        s += w + COMMAND_SPACE
    return s.strip()


def parse(pathobj):
    global DRILL_RETRACT_MODE
    global MOTION_MODE
    global CURRENT_X
    global CURRENT_Y
    global CURRENT_Z

    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"

    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "U",
        "V",
        "W",
        "I",
        "J",
        "K",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "P",
    ]

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        if OUTPUT_COMMENTS:
            out += linenumber() + "(Compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out

    else:  # Parsing simple path
        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        if OUTPUT_COMMENTS and OUTPUT_PATH:
            out += linenumber() + "(Path: " + pathobj.Label + ")\n"

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:
            outlist = []
            command = c.Name
            outlist.append(command)
            # Debug:
            # print('pathobj.Path.Commands:', c)

            # If modal is True, delete duplicate commands:
            if MODAL:
                if command == lastcommand:
                    outlist.pop(0)

            # Add the remaining parameters in order:
            for param in params:
                if param in c.Parameters:
                    if param == "F":
                        if command not in RAPID_MOVES:
                            feedRate = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                            if feedRate.getValueAs(UNIT_FEED_FORMAT) > 0.0:
                                outlist.append(
                                    param
                                    + format(
                                        float(feedRate.getValueAs(UNIT_FEED_FORMAT)),
                                        precision_string,
                                    )
                                )
                    elif param in ["T", "H", "D", "S", "P", "L"]:
                        outlist.append(param + str(c.Parameters[param]))
                    elif param in ["A", "B", "C"]:
                        outlist.append(param + format(c.Parameters[param], precision_string))
                    # [X, Y, Z, U, V, W, I, J, K, R, Q]
                    else:
                        pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                        outlist.append(
                            param + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string)
                        )

            # Store the latest command:
            lastcommand = command

            # Capture the current position for subsequent calculations:
            if command in MOTION_COMMANDS:
                if "X" in c.Parameters:
                    CURRENT_X = Units.Quantity(c.Parameters["X"], FreeCAD.Units.Length)
                if "Y" in c.Parameters:
                    CURRENT_Y = Units.Quantity(c.Parameters["Y"], FreeCAD.Units.Length)
                if "Z" in c.Parameters:
                    CURRENT_Z = Units.Quantity(c.Parameters["Z"], FreeCAD.Units.Length)

            if command in ("G98", "G99"):
                DRILL_RETRACT_MODE = command

            if TRANSLATE_DRILL_CYCLES:
                if command in ("G81", "G82", "G83"):
                    out += drill_translate(outlist, command, c.Parameters)
                    # Erase the line just translated:
                    outlist = []

            if SPINDLE_WAIT > 0:
                if command in ("M3", "M03", "M4", "M04"):
                    out += linenumber() + format_outlist(outlist) + "\n"
                    # Marlin: P for milliseconds, S for seconds, change P to S
                    out += linenumber()
                    out += format_outlist(["G4", "S%s" % SPINDLE_WAIT])
                    out += "\n"
                    outlist = []

            # Check for Tool Change:
            if command in ("M6", "M06"):
                if OUTPUT_COMMENTS:
                    out += linenumber() + "(Begin toolchange)\n"
                if OUTPUT_TOOL_CHANGE:
                    for line in TOOL_CHANGE.splitlines(True):
                        out += linenumber() + line
                if not OUTPUT_TOOL_CHANGE and OUTPUT_COMMENTS:
                    outlist[0] = "(" + outlist[0]
                    outlist[-1] = outlist[-1] + ")"
                if not OUTPUT_TOOL_CHANGE and not OUTPUT_COMMENTS:
                    outlist = []

            if command == "message":
                if OUTPUT_COMMENTS:
                    outlist.pop(0)  # remove the command
                else:
                    out = []

            if command in SUPPRESS_COMMANDS:
                outlist[0] = "(" + outlist[0]
                outlist[-1] = outlist[-1] + ")"

            # Remove embedded comments:
            if not OUTPUT_COMMENTS:
                tmplist = []
                list_index = 0
                while list_index < len(outlist):
                    left_index = outlist[list_index].find("(")
                    if left_index == -1:  # Not a comment
                        tmplist.append(outlist[list_index])
                    else:  # This line contains a comment, and possibly more
                        right_index = outlist[list_index].find(")")
                        comment_area = outlist[list_index][left_index : right_index + 1]
                        line_minus_comment = outlist[list_index].replace(comment_area, "").strip()
                        if line_minus_comment:
                            # Line contained more than just a comment
                            tmplist.append(line_minus_comment)
                    list_index += 1
                # Done removing comments
                outlist = tmplist

            # Prepend a line number and append a newline
            if len(outlist) >= 1:
                out += linenumber() + format_outlist(outlist) + "\n"

    return out


# *****************************************************************************
# * As of Marlin 2.0.7.bugfix, canned drill cycles do not exist.              *
# * The following code converts FreeCAD's canned drill cycles into            *
# * gcode that Marlin can use.                                                *
# *****************************************************************************
def drill_translate(outlist, cmd, params):
    global DRILL_RETRACT_MODE
    global MOTION_MODE
    global CURRENT_X
    global CURRENT_Y
    global CURRENT_Z
    global UNITS
    global UNIT_FORMAT
    global UNIT_FEED_FORMAT

    class Drill:  # Using a class is necessary for the nested functions.
        gcode = ""

    strFormat = "." + str(PRECISION) + "f"

    if OUTPUT_COMMENTS:  # Comment the original command
        outlist[0] = "(" + outlist[0]
        outlist[-1] = outlist[-1] + ")"
        Drill.gcode += linenumber() + format_outlist(outlist) + "\n"

    # Cycle conversion only converts the cycles in the XY plane (G17).
    # --> ZX (G18) and YZ (G19) planes produce false gcode.
    drill_X = Units.Quantity(params["X"], FreeCAD.Units.Length)
    drill_Y = Units.Quantity(params["Y"], FreeCAD.Units.Length)
    drill_Z = Units.Quantity(params["Z"], FreeCAD.Units.Length)
    drill_R = Units.Quantity(params["R"], FreeCAD.Units.Length)
    drill_F = Units.Quantity(params["F"], FreeCAD.Units.Velocity)
    if cmd == "G82":
        drill_DwellTime = params["P"]
    elif cmd == "G83":
        drill_Step = Units.Quantity(params["Q"], FreeCAD.Units.Length)

    # R less than Z is error
    if drill_R < drill_Z:
        Drill.gcode += linenumber() + "(drill cycle error: R less than Z )\n"
        return Drill.gcode

    # Z height to retract to when drill cycle is done:
    if DRILL_RETRACT_MODE == "G98" and CURRENT_Z > drill_R:
        RETRACT_Z = CURRENT_Z
    else:
        RETRACT_Z = drill_R

    # Z motion nested functions:
    def rapid_Z_to(new_Z):
        Drill.gcode += linenumber() + "G0 Z"
        Drill.gcode += format(float(new_Z.getValueAs(UNIT_FORMAT)), strFormat) + "\n"

    def feed_Z_to(new_Z):
        Drill.gcode += linenumber() + "G1 Z"
        Drill.gcode += format(float(new_Z.getValueAs(UNIT_FORMAT)), strFormat) + " F"
        Drill.gcode += format(float(drill_F.getValueAs(UNIT_FEED_FORMAT)), ".2f") + "\n"

    # Make sure that Z is not below RETRACT_Z:
    if CURRENT_Z < RETRACT_Z:
        rapid_Z_to(RETRACT_Z)

    # Rapid to hole position XY:
    Drill.gcode += linenumber() + "G0 X"
    Drill.gcode += format(float(drill_X.getValueAs(UNIT_FORMAT)), strFormat) + " Y"
    Drill.gcode += format(float(drill_Y.getValueAs(UNIT_FORMAT)), strFormat) + "\n"

    # Rapid to R:
    rapid_Z_to(drill_R)

    # *************************************************************************
    # * Drill cycles:                                                         *
    # * G80 Cancel the drill cycle                                            *
    # * G81 Drill full depth in one pass                                      *
    # * G82 Drill full depth in one pass, and pause at the bottom             *
    # * G83 Drill in pecks, raising the drill to R height after each peck     *
    # * In preparation for a rapid to the next hole position:                 *
    # * G98 After the hole has been drilled, retract to the initial Z value   *
    # * G99 After the hole has been drilled, retract to R height              *
    # * Select G99 only if safe to move from hole to hole at the R height     *
    # *************************************************************************
    if cmd in ("G81", "G82"):
        feed_Z_to(drill_Z)  # Drill hole in one step
        if cmd == "G82":  # Dwell time delay at the bottom of the hole
            Drill.gcode += linenumber() + "G4 S" + str(drill_DwellTime) + "\n"
            # Marlin uses P for milliseconds, S for seconds, change P to S

    elif cmd == "G83":  # Peck drill cycle:
        chip_Space = drill_Step * 0.5
        next_Stop_Z = drill_R - drill_Step
        while next_Stop_Z >= drill_Z:
            feed_Z_to(next_Stop_Z)  # Drill one peck of depth

            # Set next depth, next_Stop_Z is still at the current hole depth
            if (next_Stop_Z - drill_Step) >= drill_Z:
                # Rapid up to clear chips:
                rapid_Z_to(drill_R)
                # Rapid down to just above last peck depth:
                rapid_Z_to(next_Stop_Z + chip_Space)
                # Update next_Stop_Z to next depth:
                next_Stop_Z -= drill_Step
            elif next_Stop_Z == drill_Z:
                break  # Done
            else:  # More to drill, but less than drill_Step
                # Rapid up to clear chips:
                rapid_Z_to(drill_R)
                # Rapid down to just above last peck depth:
                rapid_Z_to(next_Stop_Z + chip_Space)
                # Dril remainder of the hole depth:
                feed_Z_to(drill_Z)
                break  # Done
    rapid_Z_to(RETRACT_Z)  # Done, retract the drill

    return Drill.gcode


# print(__name__ + ': GCode postprocessor loaded.')

# PEP8 format passed using: http://pep8online.com/, which primarily covers
# indentation and line length. Some other aspects of PEP8 which have not
# been applied yet may be applied in future updates.



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import os
import FreeCAD as App

from PySide.QtCore import QT_TRANSLATE_NOOP

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets

import UtilsAssembly
import Assembly_rc


__title__ = "Assembly Command to Solve Assembly"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


class CommandSolveAssembly:
    def __init__(self):
        pass

    def GetResources(self):

        return {
            "Pixmap": "Assembly_SolveAssembly",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_SolveAssembly", "Solve Assembly"),
            "Accel": "Z",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_SolveAssembly",
                "Solves the currently active assembly.",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        return UtilsAssembly.isAssemblyCommandActive()

    def Activated(self):
        assembly = UtilsAssembly.activeAssembly()
        if not assembly:
            return

        App.setActiveTransaction("Solve assembly")
        assembly.recompute(True)
        App.closeActiveTransaction()


if App.GuiUp:
    Gui.addCommand("Assembly_SolveAssembly", CommandSolveAssembly())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Label object."""
## @package label
# \ingroup draftobjects
# \brief Provides the object code for the Label object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from FreeCAD import Units as U
from draftobjects.draft_annotation import DraftAnnotation
from draftutils import gui_utils
from draftutils.messages import _log
from draftutils.translate import translate


class Label(DraftAnnotation):
    """The Draft Label object."""

    def __init__(self, obj):
        obj.Proxy = self
        self.Type = "Label"
        self.set_properties(obj)

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        self.set_target_properties(obj)
        self.set_leader_properties(obj)
        self.set_label_properties(obj)

    def set_target_properties(self, obj):
        """Set position properties only if they don't exist."""
        properties = obj.PropertiesList

        if "TargetPoint" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The position of the tip of the leader "
                "line.\n"
                "This point can be decorated "
                "with an arrow or another symbol.",
            )
            obj.addProperty("App::PropertyVector", "TargetPoint", "Target", _tip, locked=True)
            obj.TargetPoint = App.Vector(2, -1, 0)

        if "Target" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Object, and optionally subelement, "
                "whose properties will be displayed\n"
                "as 'Text', depending on 'Label Type'.\n"
                "\n"
                "'Target' won't be used "
                "if 'Label Type' is set to 'Custom'.",
            )
            obj.addProperty("App::PropertyLinkSub", "Target", "Target", _tip, locked=True)
            obj.Target = None

    def set_leader_properties(self, obj):
        """Set leader properties only if they don't exist."""
        properties = obj.PropertiesList

        if "Points" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The list of points defining the leader "
                "line; normally a list of three points.\n"
                "\n"
                "The first point should be the position "
                "of the text, that is, the 'Placement',\n"
                "and the last point should be "
                "the tip of the line, that is, "
                "the 'Target Point'.\n"
                "The middle point is calculated "
                "automatically depending on the chosen\n"
                "'Straight Direction' "
                "and the 'Straight Distance' value "
                "and sign.\n"
                "\n"
                "If 'Straight Direction' is set to "
                "'Custom', the 'Points' property\n"
                "can be set as a list "
                "of arbitrary points.",
            )
            obj.addProperty("App::PropertyVectorList", "Points", "Leader", _tip, locked=True)
            obj.Points = []

        if "StraightDirection" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The direction of the straight segment "
                "of the leader line.\n"
                "\n"
                "If 'Custom' is chosen, the points "
                "of the leader can be specified by\n"
                "assigning a custom list "
                "to the 'Points' attribute.",
            )
            obj.addProperty(
                "App::PropertyEnumeration", "StraightDirection", "Leader", _tip, locked=True
            )
            obj.StraightDirection = ["Horizontal", "Vertical", "Custom"]

        if "StraightDistance" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The length of the straight segment "
                "of the leader line.\n"
                "\n"
                "This is an oriented distance; "
                "if it is negative, the line will "
                "be drawn\n"
                "to the left or below the 'Text', "
                "otherwise to the right or above it,\n"
                "depending on the value of "
                "'Straight Direction'.",
            )
            obj.addProperty(
                "App::PropertyDistance", "StraightDistance", "Leader", _tip, locked=True
            )
            obj.StraightDistance = 1

    def set_label_properties(self, obj):
        """Set label properties only if they don't exist."""
        properties = obj.PropertiesList

        if "Placement" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The placement of the 'Text' element " "in 3D space"
            )
            obj.addProperty("App::PropertyPlacement", "Placement", "Label", _tip, locked=True)
            obj.Placement = App.Placement()

        if "CustomText" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The text to display when 'Label Type' " "is set to 'Custom'"
            )
            obj.addProperty("App::PropertyStringList", "CustomText", "Label", _tip, locked=True)
            obj.CustomText = "Label"

        if "Text" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The text displayed by this label.\n"
                "\n"
                "This property is read-only, as the "
                "final text depends on 'Label Type',\n"
                "and the object defined in 'Target'.\n"
                "The 'Custom Text' is displayed only "
                "if 'Label Type' is set to 'Custom'.",
            )
            obj.addProperty("App::PropertyStringList", "Text", "Label", _tip, locked=True)
            obj.setEditorMode("Text", 1)  # Read only

        # TODO: maybe here we can define a second and third 'label type'
        # properties, so that the final displayed text is either
        # the first type, or the combination of two or three types,
        # if they are available.
        # The current system has some labels combined, but these combinations
        # are hard coded. By considering multiple properties, we could produce
        # arbitrary combinations of labels.
        # This would also require updating the `return_info` function
        # to handle any combination that we want.
        if "LabelType" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The type of information displayed "
                "by this label.\n"
                "\n"
                "If 'Custom' is chosen, the contents of "
                "'Custom Text' will be used.\n"
                "For other types, the string will be "
                "calculated automatically from the "
                "object defined in 'Target'.\n"
                "'Tag' and 'Material' only work "
                "for objects that have these properties, "
                "like BIM objects.\n"
                "\n"
                "For 'Position', 'Length', and 'Area' "
                "these properties will be extracted "
                "from the main object in 'Target',\n"
                "or from the subelement "
                "'VertexN', 'EdgeN', or 'FaceN', "
                "respectively, if it is specified.",
            )
            obj.addProperty("App::PropertyEnumeration", "LabelType", "Label", _tip, locked=True)
            obj.LabelType = get_label_types()

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored."""
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_label", vp_class="ViewProviderLabel")

        vobj = getattr(obj, "ViewObject", None)
        if vobj is None:
            return

        if not hasattr(vobj, "FontName") or not hasattr(vobj, "FontSize"):
            self.update_properties_0v21(obj, vobj)

    def loads(self, state):
        self.Type = "Label"

    def update_properties_0v21(self, obj, vobj):
        """Update view properties."""
        old_fontname = vobj.TextFont
        old_fontsize = vobj.TextSize
        vobj.removeProperty("TextFont")
        vobj.removeProperty("TextSize")

        vobj.Proxy.set_text_properties(vobj, vobj.PropertiesList)
        vobj.FontName = old_fontname
        vobj.FontSize = old_fontsize
        # The DisplayMode is updated automatically but the new values are
        # switched: "2D text" becomes "World" and "3D text" becomes "Screen".
        # It should be the other way around:
        vobj.DisplayMode = "World" if vobj.DisplayMode == "Screen" else "Screen"
        _log("v0.21, " + obj.Name + ", renamed view property 'TextFont' to 'FontName'")
        _log("v0.21, " + obj.Name + ", renamed view property 'TextSize' to 'FontSize'")
        _log("v0.21, " + obj.Name + ", renamed 'DisplayMode' options to 'World/Screen'")

    def onChanged(self, obj, prop):
        """Execute when a property is changed."""
        self.show_and_hide(obj, prop)

    def show_and_hide(self, obj, prop):
        """Show and hide the properties depending on the touched property."""
        # The minus sign removes the Hidden property (show)
        if prop == "LabelType":
            if obj.LabelType != "Custom":
                obj.setPropertyStatus("CustomText", "Hidden")
                obj.setPropertyStatus("Target", "-Hidden")
            else:
                obj.setPropertyStatus("CustomText", "-Hidden")
                obj.setPropertyStatus("Target", "Hidden")

    def execute(self, obj):
        """Execute when the object is created or recomputed."""
        if obj.StraightDirection != "Custom":
            p1 = obj.Placement.Base
            if obj.StraightDirection == "Horizontal":
                p2 = App.Vector(obj.StraightDistance.Value, 0, 0)
            elif obj.StraightDirection == "Vertical":
                p2 = App.Vector(0, obj.StraightDistance.Value, 0)

            p2 = obj.Placement.multVec(p2)
            # p3 = obj.Placement.multVec(obj.TargetPoint)
            p3 = obj.TargetPoint
            obj.Points = [p1, p2, p3]
        else:
            # If StraightDirection is 'Custom'
            # we can draw the leader line manually by specifying
            # any number of vectors in the Points property.
            # The first point should indicate the position of the text label,
            # while the last one should be 'TargetPoint'
            # obj.Points = [p1, p2, p3, p4, ...]
            #
            # The drawing of the line is done in the viewprovider
            #
            # However, as soon as StraightDirection is changed to
            # 'Horizontal' or 'Vertical' this custom list of points
            # will be overwritten
            pass

        if obj.LabelType == "Custom":
            if obj.CustomText:
                obj.Text = obj.CustomText

        elif obj.Target and obj.Target[0]:
            target = obj.Target[0]
            sub_list = obj.Target[1]
            typ = obj.LabelType

            # The sublist may be empty so we test it first
            subelement = sub_list[0] if sub_list else None
            obj.Text = return_info(target, typ, subelement)

        else:
            obj.Text = [translate("draft", "No Target")]


# Alias for compatibility with v0.18 and earlier
DraftLabel = Label


def get_label_types():
    return [
        QT_TRANSLATE_NOOP("Draft", "Custom"),
        QT_TRANSLATE_NOOP("Draft", "Name"),
        QT_TRANSLATE_NOOP("Draft", "Label"),
        QT_TRANSLATE_NOOP("Draft", "Position"),
        QT_TRANSLATE_NOOP("Draft", "Length"),
        QT_TRANSLATE_NOOP("Draft", "Area"),
        QT_TRANSLATE_NOOP("Draft", "Volume"),
        QT_TRANSLATE_NOOP("Draft", "Tag"),
        QT_TRANSLATE_NOOP("Draft", "Material"),
        QT_TRANSLATE_NOOP("Draft", "Label + Position"),
        QT_TRANSLATE_NOOP("Draft", "Label + Length"),
        QT_TRANSLATE_NOOP("Draft", "Label + Area"),
        QT_TRANSLATE_NOOP("Draft", "Label + Volume"),
        QT_TRANSLATE_NOOP("Draft", "Label + Material"),
    ]


def return_info(target, typ, subelement=None):
    """Return the text list from the target and the given type.

    Parameters
    ----------
    target: Part::Feature
        The object targeted by the label.

    typ: str
        It is the type of information that we want to extract.

    subelement: str, optional
        A string indicating a subelement of the `target`;
        it could be `'VertexN'`, `'EdgeN'`, or `'FaceN'`,
        where `'N'` is a number that starts from `1` up to the maximum
        number of subelements in that target.
    """
    # print(obj, target, typ, subelement)

    if typ == "Name":
        return _get_name(target)

    if typ == "Label":
        return _get_label(target)

    if typ == "Tag":
        return _get_tag(target)

    if typ == "Material":
        return _get_material(target)

    if typ == "Label + Material":
        return _get_label(target) + _get_material(target)

    if typ == "Position":
        return _get_position(target, subelement)

    if typ == "Label + Position":
        return _get_label(target) + _get_position(target, subelement)

    if typ == "Length":
        return _get_length(target, subelement)

    if typ == "Label + Length":
        return _get_label(target) + _get_length(target, subelement)

    if typ == "Area":
        return _get_area(target, subelement)

    if typ == "Label + Area":
        return _get_label(target) + _get_area(target, subelement)

    if typ == "Volume":
        return _get_volume(target, subelement)

    if typ == "Label + Volume":
        return _get_label(target) + _get_volume(target, subelement)

    return [translate("draft", "Invalid label type")]


def _get_name(target):
    return [target.Name]


def _get_label(target):
    return [target.Label]


def _get_tag(target):
    if hasattr(target, "Tag"):
        return [target.Tag]
    else:
        return [translate("draft", "Tag not available for object")]


def _get_material(target):
    if hasattr(target, "Material") and hasattr(target.Material, "Label"):
        return [target.Material.Label]
    else:
        return [translate("draft", "Material not available for object")]


def _get_position(target, subelement):
    point = None
    if subelement is not None:
        if "Vertex" in subelement:
            point = target.Shape.Vertexes[int(subelement[6:]) - 1].Point
    else:
        point = target.Placement.Base
    if point is None:
        return [translate("draft", "Position not available for (sub)object")]
    return [U.Quantity(x, U.Length).UserString for x in tuple(point)]


def _get_length(target, subelement):
    length = None
    if subelement is not None:
        if "Edge" in subelement:
            length = target.Shape.Edges[int(subelement[4:]) - 1].Length
        elif "Face" in subelement:
            length = target.Shape.Faces[int(subelement[4:]) - 1].Length
    elif hasattr(target, "Length"):
        length = target.Length
    elif hasattr(target, "Shape") and hasattr(target.Shape, "Length"):
        length = target.Shape.Length
    if length is None:
        return [translate("draft", "Length not available for (sub)object")]
    return [U.Quantity(length, U.Length).UserString]


def _get_area(target, subelement):
    area = None
    if subelement is not None:
        if "Face" in subelement:
            area = target.Shape.Faces[int(subelement[4:]) - 1].Area
    elif hasattr(target, "Area"):
        area = target.Area
    elif hasattr(target, "Shape") and hasattr(target.Shape, "Area"):
        area = target.Shape.Area
    if area is None:
        return [translate("draft", "Area not available for (sub)object")]
    return [U.Quantity(area, U.Area).UserString.replace("^2", "²")]


def _get_volume(target, subelement):
    volume = None
    if subelement is not None:
        pass
    elif hasattr(target, "Volume"):
        volume = target.Volume
    elif hasattr(target, "Shape") and hasattr(target.Shape, "Volume"):
        volume = target.Shape.Volume
    if volume is None:
        return [translate("draft", "Volume not available for (sub)object")]
    return [U.Quantity(volume, U.Volume).UserString.replace("^3", "³")]


## @}



# ***************************************************************************
# *   Copyright (c) 2015 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Common FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import unittest

import FreeCAD

import ObjectsFem
from . import support_utils as testtools
from .support_utils import fcc_print


class TestFemCommon(unittest.TestCase):
    fcc_print("import TestFemCommon")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestFemCommon tests {2}\n{0}".format(100 * "*", 10 * "*", 61 * "*")
        )

    # ********************************************************************************************
    def test_adding_refshaps(self):
        doc = self.document
        slab = doc.addObject("Part::Plane", "Face")
        slab.Length = 500.00
        slab.Width = 500.00
        cf = ObjectsFem.makeConstraintFixed(doc)
        ref_eles = []
        # FreeCAD list property doesn't seem to support append,
        # thus we need a workaround
        # which on many elements is even much faster
        for i, face in enumerate(slab.Shape.Edges):
            ref_eles.append("Edge%d" % (i + 1))
        cf.References = [(slab, ref_eles)]
        doc.recompute()
        expected_reflist = [(slab, ("Edge1", "Edge2", "Edge3", "Edge4"))]
        assert_err_message = (
            "Adding reference shapes did not result in expected list {} != {}".format(
                cf.References, expected_reflist
            )
        )
        self.assertEqual(cf.References, expected_reflist, assert_err_message)

    # ********************************************************************************************
    def test_pyimport_all_FEM_modules(self):
        # we're going to try to import all python modules from FreeCAD FEM
        pymodules = []

        # collect all Python modules in FEM
        pymodules += testtools.collect_python_modules("")  # FEM main dir
        pymodules += testtools.collect_python_modules("femexamples")
        pymodules += testtools.collect_python_modules("feminout")
        pymodules += testtools.collect_python_modules("femmesh")
        pymodules += testtools.collect_python_modules("femobjects")
        pymodules += testtools.collect_python_modules("femresult")
        pymodules += testtools.collect_python_modules("femtest")
        pymodules += testtools.collect_python_modules("femtools")
        pymodules += testtools.collect_python_modules("femsolver")
        # TODO test with join on Windows, the use of os.path.join
        # in the following code seems to create problems on Windows OS
        pymodules += testtools.collect_python_modules("femsolver/elmer")
        pymodules += testtools.collect_python_modules("femsolver/elmer/equations")
        pymodules += testtools.collect_python_modules("femsolver/z88")
        pymodules += testtools.collect_python_modules("femsolver/calculix")
        if FreeCAD.GuiUp:
            pymodules += testtools.collect_python_modules("femcommands")
            pymodules += testtools.collect_python_modules("femguiobjects")
            pymodules += testtools.collect_python_modules("femguiutils")
            pymodules += testtools.collect_python_modules("femtaskpanels")
            pymodules += testtools.collect_python_modules("femviewprovider")

        # import all collected modules
        # fcc_print(pymodules)
        for mod in pymodules:
            if (
                mod == "femsolver.solver_taskpanel"
                or mod == "femexamples.examplesgui"
                or mod == "TestFemGui"
            ) and not FreeCAD.GuiUp:
                continue

            fcc_print(f"Try importing {mod} ...")
            try:
                im = __import__(f"{mod}")
            except ImportError as e:
                # check if it is a VTK module that is missing, because maybe we should not need it
                if "vtkmodules" in e.name and not "BUILD_FEM_VTK_PYTHON" in FreeCAD.__cmake__:
                    im = True
                else:
                    im = False
            if not im:
                # to get an error message what was going wrong
                __import__(f"{mod}")
            self.assertTrue(im, f"Problem importing {mod}")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (C) 2020 Stefano Chiaro <stefano.chiaro@yahoo.com>          *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Lesser General Public            *
# *   License as published by the Free Software Foundation; either          *
# *   version 2.1 of the License, or (at your option) any later version.    *
# *                                                                         *
# *   This library is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with this library; if not, write to the Free Software   *
# *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
# *   02110-1301  USA                                                       *
# ***************************************************************************

# HEDENHAIN Post-Processor for FreeCAD

import argparse
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
import Path
import PathScripts
import shlex
import math
from builtins import open as pyopen

# **************************************************************************#
#                   USER  EDITABLE  STUFF  HERE                            #
#                                                                          #
#     THESE VALUES SHOULD BE CHANGED TO FIT MACHINE TYPE AND LANGUAGE      #

MACHINE_SKIP_PARAMS = False  # Print R F M values
# possible values:
#                    'True'    Skip to print a parameter if already active
#                    'False'   Print parameter on ALL line
# Old machines need these values on every lines

MACHINE_USE_FMAX = False  # Usage of FMAX
# possible values:
#                 'True'   Print FMAX
#                 'False'  Print the value set on FEED_MAX_SPEED
# Old machines don't accept FMAX and need a feed value

FEED_MAX_SPEED = 8000  # Max machine speed for FMAX
# possible values:
# integer >= 0

AXIS_DECIMALS = 3  # machine axis precision
# possible values:
# integer >= 0

FEED_DECIMALS = 0  # machine feed precision
# possible values:
# integer >= 0

SPINDLE_DECIMALS = 0  # machine spindle precision
# possible values:
# integer >= 0

FIRST_LBL = 1  # first LBL number for LBLIZE function
# possible values:
# integer >= 0

#                   TEMPLATES FOR CYCLE DEFINITION                         #

MACHINE_CYCLE_DEF = {
    1: "CYCL DEF 1.0 FORATURA PROF."
    + "\n"
    + "CYCL DEF 1.1 DIST"
    + "{DIST}\n"
    + "CYCL DEF 1.2 PROF"
    + "{DEPTH}\n"
    + "CYCL DEF 1.3 INCR"
    + "{INCR}\n"
    + "CYCL DEF 1.4 SOSTA"
    + "{DWELL}\n"
    + "CYCL DEF 1.5 F"
    + "{FEED}",
    2: "",
    7: "",
}

#                      OPTIONAL COMPUTING                                  #

SOLVE_COMPENSATION_ACTIVE = False  # Use the internal path compensation
# possible values:
#                          'True'   Try to solve compensation
#                          'False'  Use original FreeCAD path
# try to use RL and RR to get the real path where possible

SHOW_EDITOR = True  # Open the editor
# possible values:
#            'True'   before the file is written it is shown for inspection
#            'False'  the file is written directly

SKIP_WARNS = False  # Skip post-processor warnings
# possible values:
#            'True'   never prompt warning from post-processing problems
#            'False'  prompt active

#            STANDARD HEIDENHAIN VALUES FOR POSTPROCESSOR                  #

UNITS = "MM"  # post-processor units
# possible values:
#       'INCH'  for inches
#       'MM'    for metric units
# actually FreeCAD use only metric values

LINENUMBERS = True  # line numbers
# possible values:
#            'True'   Add line numbers (Standard)
#            'False'  No line numbers

STARTLINENR = 0  # first line number used
# possible values:
# any integer value >= 0 (Standard is 0)

LINENUMBER_INCREMENT = 1  # line number increment
# possible values:
# any integer value > 0 (Standard is 1)

#                POSTPROCESSOR VARIABLES AND CODE                          #

# **************************************************************************#
#  don't edit the stuff below this line unless you know what you're doing  #
# **************************************************************************#

#                   GCODE VARIABLES AND FUNCTIONS                          #

POSTGCODE = []  # Output string array
G_FUNCTION_STORE = {
    "G90": False,
    "G91": False,
    "G98": False,
    "G99": False,
    "G81": False,
    "G82": False,
    "G83": False,
}

#                     HEIDENHAIN MACHINE PARAMETERS                        #

MACHINE_WORK_AXIS = 2  # 0=X ; 1=Y ; 2=Z usually Z
MACHINE_SPINDLE_DIRECTION = 3  # CW = 3 ; CCW = 4
MACHINE_LAST_POSITION = {  # axis initial values to overwrite
    "X": 99999,
    "Y": 99999,
    "Z": 99999,
    "A": 99999,
    "B": 99999,
    "C": 99999,
}
MACHINE_LAST_CENTER = {  # CC initial values to overwrite
    "X": 99999,
    "Y": 99999,
    "Z": 99999,
}
MACHINE_TRASL_ROT = [0, 0, 0, 0]  # [X, Y, Z , Angle] !not implemented
MACHINE_STORED_PARAMS = ["", -1, ""]  # Store R F M parameter to skip

#                  POSTPROCESSOR VARIABLES STORAGE                         #

STORED_COMPENSATED_OBJ = ()  # Store a copy of compensated path
STORED_CANNED_PARAMS = {  # Store canned cycles for match
    "DIST": 99999,
    "DEPTH": 99999,
    "INCR": 99999,
    "DWELL": 99999,
    "FEED": 99999,
}
STORED_LBL = []  # Store array of LBL for match

#                  POSTPROCESSOR SPECIAL VARIABLES                         #

COMPENSATION_DIFF_STATUS = [False, True]  # Check compensation, Check Diff
LBLIZE_ACTIVE = False  # Check if search for LBL
LBLIZE_STAUS = False  # Activated by path type
LBLIZE_PATH_LEVEL = 99999  # Save milling level of actual LBL

#                   END OF POSTPROCESSOR VARIABLES                         #
# **************************************************************************#

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-gcode fragment outputted by a Path object, and output
real GCode suitable for a heidenhain 3 axis mill. This postprocessor, once placed
in the appropriate Path/Tool folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import heidenhain_post
heidenhain.export(object,"/path/to/file.ncc","")
"""

parser = argparse.ArgumentParser(prog="heidenhain", add_help=False)
parser.add_argument(
    "--skip-params",
    action="store_true",
    help="suppress R F M parameters where already stored",
)
parser.add_argument(
    "--use-fmax",
    action="store_true",
    help="suppress feedrate and use FMAX instead with rapid movements",
)
parser.add_argument(
    "--fmax-value", default="8000", help="feedrate to use instead of FMAX, default=8000"
)
parser.add_argument(
    "--axis-decimals", default="3", help="number of digits of axis precision, default=3"
)
parser.add_argument(
    "--feed-decimals",
    default="0",
    help="number of digits of feedrate precision, default=0",
)
parser.add_argument(
    "--spindle-decimals",
    default="0",
    help="number of digits of spindle precision, default=0",
)
parser.add_argument(
    "--solve-comp",
    action="store_true",
    help="try to get RL or RR real path where compensation is active",
)
parser.add_argument(
    "--solve-lbl",
    action="store_true",
    help="try to replace repetitive movements with LBL",
)
parser.add_argument("--first-lbl", default="1", help="change the first LBL number, default=1")
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--no-warns", action="store_true", help="don't pop up post-processor warnings")

TOOLTIP_ARGS = parser.format_help()


def processArguments(argstring):
    global MACHINE_SKIP_PARAMS
    global MACHINE_USE_FMAX
    global FEED_MAX_SPEED
    global AXIS_DECIMALS
    global FEED_DECIMALS
    global SPINDLE_DECIMALS
    global SOLVE_COMPENSATION_ACTIVE
    global LBLIZE_ACTIVE
    global FIRST_LBL
    global SHOW_EDITOR
    global SKIP_WARNS

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.skip_params:
            MACHINE_SKIP_PARAMS = True
        if args.use_fmax:
            MACHINE_USE_FMAX = True
        if args.fmax_value:
            FEED_MAX_SPEED = float(args.fmax_value)
        if args.axis_decimals:
            AXIS_DECIMALS = int(args.axis_decimals)
        if args.feed_decimals:
            FEED_DECIMALS = int(args.feed_decimals)
        if args.spindle_decimals:
            SPINDLE_DECIMALS = int(args.spindle_decimals)
        if args.solve_comp:
            SOLVE_COMPENSATION_ACTIVE = True
        if args.solve_lbl:
            LBLIZE_ACTIVE = True
        if args.first_lbl:
            FIRST_LBL = int(args.first_lbl)
        if args.no_show_editor:
            SHOW_EDITOR = False
        if args.no_warns:
            SKIP_WARNS = True
    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None
    global UNITS
    global POSTGCODE
    global G_FUNCTION_STORE
    global MACHINE_WORK_AXIS
    global MACHINE_LAST_POSITION
    global MACHINE_TRASL_ROT
    global STORED_COMPENSATED_OBJ
    global COMPENSATION_DIFF_STATUS
    global LBLIZE_STAUS

    Object_Kind = None
    Feed = 0
    Spindle_Active = False
    Compensation = "0"
    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "K",
        "F",
        "H",
        "S",
        "T",
        "Q",
        "R",
        "L",
    ]

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return

    POSTGCODE.append(HEIDEN_Begin(objectslist))  # add header

    for obj in objectslist:
        Cmd_Count = 0  # command line number
        LBLIZE_STAUS = False

        if not hasattr(obj, "Proxy"):
            continue
        # useful to get idea of object kind
        if isinstance(obj.Proxy, Path.Tool.Controller.ToolController):
            Object_Kind = "TOOL"
            # like we go to change tool position
            MACHINE_LAST_POSITION["X"] = 99999
            MACHINE_LAST_POSITION["Y"] = 99999
            MACHINE_LAST_POSITION["Z"] = 99999
        elif isinstance(obj.Proxy, Path.Op.ProfileEdges.ObjectProfile):
            Object_Kind = "PROFILE"
            if LBLIZE_ACTIVE:
                LBLIZE_STAUS = True
        elif isinstance(obj.Proxy, Path.Op.MillFace.ObjectFace):
            Object_Kind = "FACE"
            if LBLIZE_ACTIVE:
                LBLIZE_STAUS = True
        elif isinstance(obj.Proxy, Path.Op.Helix.ObjectHelix):
            Object_Kind = "HELIX"

        commands = PathUtils.getPathWithPlacement(obj).Commands

        # If used compensated path, store, recompute and diff when asked
        if not hasattr(obj, "UseComp") or not SOLVE_COMPENSATION_ACTIVE:
            continue

        if not obj.UseComp:
            continue

        if hasattr(obj.Path, "Commands") and Object_Kind == "PROFILE":
            # Take a copy of compensated path
            STORED_COMPENSATED_OBJ = commands
        # Find mill compensation
        if hasattr(obj, "Side") and hasattr(obj, "Direction"):
            if obj.Side == "Outside" and obj.Direction == "CW":
                Compensation = "L"
            elif obj.Side == "Outside" and obj.Direction == "CCW":
                Compensation = "R"
            elif obj.Side != "Outside" and obj.Direction == "CW":
                Compensation = "R"
            else:
                Compensation = "L"
            # set obj.UseComp to false and recompute() to get uncompensated path
            obj.UseComp = False
            obj.recompute()
            commands = PathUtils.getPathWithPlacement(obj).Commands
            # small edges could be skipped and movements joints can add edges
            NameStr = ""
            if hasattr(obj, "Label"):
                NameStr = str(obj.Label)
            if len(commands) != len(STORED_COMPENSATED_OBJ):
                # not same number of edges
                obj.UseComp = True
                obj.recompute()
                commands = PathUtils.getPathWithPlacement(obj).Commands
                POSTGCODE.append("; MISSING EDGES UNABLE TO GET COMPENSATION")
                if not SKIP_WARNS:
                    (
                        PostUtils.editor(
                            "--solve-comp command ACTIVE\n\n"
                            + "UNABLE to solve "
                            + NameStr
                            + " compensation\n\n"
                            + "Some edges are missing\n"
                            + "try to change Join Type to Miter or Square\n"
                            + "try to use a smaller Tool Diameter\n"
                            + "Internal Path could have too small corners\n\n"
                            + "use --no-warns to not prompt this message"
                        )
                    )
            else:
                if not SKIP_WARNS:
                    (
                        PostUtils.editor(
                            "--solve-comp command ACTIVE\n\n"
                            + "BE CAREFUL with solved "
                            + NameStr
                            + " compensation\n\n"
                            + "USE AT YOUR OWN RISK\n"
                            + "Simulate it before use\n"
                            + "Offset Extra ignored use DR+ on TOOL CALL\n"
                            + "Path could be different and/or give tool radius errors\n\n"
                            + "use --no-warns to not prompt this message"
                        )
                    )
                # we can try to solve compensation
                POSTGCODE.append("; COMPENSATION ACTIVE")
                COMPENSATION_DIFF_STATUS[0] = True

        for c in commands:
            Cmd_Count += 1
            command = c.Name
            if command != "G0":
                command = command.replace("G0", "G")  # normalize: G01 -> G1

            for param in params:
                if param in c.Parameters:
                    if param == "F":
                        Feed = c.Parameters["F"]

            if command == "G90":
                G_FUNCTION_STORE["G90"] = True
                G_FUNCTION_STORE["G91"] = False

            if command == "G91":
                G_FUNCTION_STORE["G91"] = True
                G_FUNCTION_STORE["G90"] = False

            if command == "G98":
                G_FUNCTION_STORE["G98"] = True
                G_FUNCTION_STORE["G99"] = False

            if command == "G99":
                G_FUNCTION_STORE["G99"] = True
                G_FUNCTION_STORE["G98"] = False

            # Rapid movement
            if command == "G0":
                Spindle_Status = ""
                if Spindle_Active == False:  # At first rapid movement we turn on spindle
                    Spindle_Status += str(MACHINE_SPINDLE_DIRECTION)  # Activate spindle
                    Spindle_Active = True
                else:  # At last rapid movement we turn off spindle
                    if Cmd_Count == len(commands):
                        Spindle_Status += "5"  # Deactivate spindle
                        Spindle_Active = False
                    else:
                        Spindle_Status += ""  # Spindle still active
                parsedElem = HEIDEN_Line(
                    c.Parameters, Compensation, Feed, True, Spindle_Status, Cmd_Count
                )
                if parsedElem is not None:
                    POSTGCODE.append(parsedElem)

            # Linear movement
            if command == "G1":
                parsedElem = HEIDEN_Line(c.Parameters, Compensation, Feed, False, "", Cmd_Count)
                if parsedElem is not None:
                    POSTGCODE.append(parsedElem)

            # Arc movement
            if command == "G2" or command == "G3":
                parsedElem = HEIDEN_Arc(
                    c.Parameters, command, Compensation, Feed, False, "", Cmd_Count
                )
                if parsedElem is not None:
                    POSTGCODE.extend(parsedElem)

            if command == "G80":  # Reset Canned Cycles
                G_FUNCTION_STORE["G81"] = False
                G_FUNCTION_STORE["G82"] = False
                G_FUNCTION_STORE["G83"] = False

            # Drilling, Dwell Drilling, Peck Drilling
            if command == "G81" or command == "G82" or command == "G83":
                parsedElem = HEIDEN_Drill(obj, c.Parameters, command, Feed)
                if parsedElem is not None:
                    POSTGCODE.extend(parsedElem)

            # Tool change
            if command == "M6":
                parsedElem = HEIDEN_ToolCall(obj)
                if parsedElem is not None:
                    POSTGCODE.append(parsedElem)

        if COMPENSATION_DIFF_STATUS[0]:  # Restore the compensation if removed
            obj.UseComp = True
            obj.recompute()
            COMPENSATION_DIFF_STATUS[0] = False

        if LBLIZE_STAUS:
            HEIDEN_LBL_Replace()
            LBLIZE_STAUS = False

    if LBLIZE_ACTIVE:
        if not SKIP_WARNS:
            (
                PostUtils.editor(
                    "--solve-lbl command ACTIVE\n\n"
                    + "BE CAREFUL with LBL replacements\n\n"
                    + "USE AT YOUR OWN RISK\n"
                    + "Simulate it before use\n"
                    + "Path could be different and/or give errors\n\n"
                    + "use --no-warns to not prompt this message"
                )
            )
    POSTGCODE.append(HEIDEN_End(objectslist))  # add footer
    Program_Out = HEIDEN_Numberize(POSTGCODE)  # add line number

    if SHOW_EDITOR:
        PostUtils.editor(Program_Out)

    gfile = pyopen(filename, "w")
    gfile.write(Program_Out)
    gfile.close()


def HEIDEN_Begin(ActualJob):  # use Label for program name
    global UNITS
    # JobParent = PathUtils.findParentJob(ActualJob[0])
    # if hasattr(JobParent, "Label"):
    #     program_id = JobParent.Label
    # else:
    #     program_id = "NEW"
    return "BEGIN PGM {}".format(UNITS)


def HEIDEN_End(ActualJob):  # use Label for program name
    global UNITS
    # JobParent = PathUtils.findParentJob(ActualJob[0])
    # if hasattr(JobParent, "Label"):
    #     program_id = JobParent.Label
    # else:
    #     program_id = "NEW"
    return "END PGM {}".format(UNITS)


# def HEIDEN_ToolDef(tool_id, tool_length, tool_radius): # old machines don't have tool table, need tooldef list
#    return "TOOL DEF  " + tool_id + " R" + "{:.3f}".format(tool_length) + " L" + "{:.3f}".format(tool_radius)


def HEIDEN_ToolCall(tool_Params):
    global MACHINE_SPINDLE_DIRECTION
    global MACHINE_WORK_AXIS
    H_Tool_Axis = ["X", "Y", "Z"]
    H_Tool_ID = "0"
    H_Tool_Speed = 0
    H_Tool_Comment = ""

    if hasattr(tool_Params, "Label"):  # use Label as tool comment
        H_Tool_Comment = tool_Params.Label
    if hasattr(tool_Params, "SpindleDir"):  # get spindle direction for this tool
        if tool_Params.SpindleDir == "Forward":
            MACHINE_SPINDLE_DIRECTION = 3
        else:
            MACHINE_SPINDLE_DIRECTION = 4
    if hasattr(tool_Params, "SpindleSpeed"):  # get tool speed for spindle
        H_Tool_Speed = tool_Params.SpindleSpeed
    if hasattr(tool_Params, "ToolNumber"):  # use ToolNumber for tool id
        H_Tool_ID = tool_Params.ToolNumber

    if H_Tool_ID == "0" and H_Tool_Speed == 0 and H_Tool_Comment == "":
        return None
    else:
        return (
            "TOOL CALL "
            + str(H_Tool_ID)
            + " "
            + H_Tool_Axis[MACHINE_WORK_AXIS]
            + HEIDEN_Format(" S", H_Tool_Speed)
            + " ;"
            + str(H_Tool_Comment)
        )


# create a linear movement
def HEIDEN_Line(line_Params, line_comp, line_feed, line_rapid, line_M_funct, Cmd_Number):
    global FEED_MAX_SPEED
    global COMPENSATION_DIFF_STATUS
    global G_FUNCTION_STORE
    global MACHINE_WORK_AXIS
    global MACHINE_LAST_POSITION
    global MACHINE_STORED_PARAMS
    global MACHINE_SKIP_PARAMS
    global MACHINE_USE_FMAX
    H_Line_New = {  # axis initial values to overwrite
        "X": 99999,
        "Y": 99999,
        "Z": 99999,
        "A": 99999,
        "B": 99999,
        "C": 99999,
    }
    H_Line = "L"
    H_Line_Params = [["X", "Y", "Z"], ["R0", line_feed, line_M_funct]]

    # check and hide duplicated axis movements, not last, update with new ones
    for i in H_Line_New:
        if G_FUNCTION_STORE["G91"]:  # incremental
            H_Line_New[i] = 0
            if i in line_Params:
                if line_Params[i] != 0 or line_M_funct != "":
                    H_Line += " I" + HEIDEN_Format(i, line_Params[i])  # print incremental
            # update to absolute position
            H_Line_New[i] = MACHINE_LAST_POSITION[i] + H_Line_New[i]
        else:  # absolute
            H_Line_New[i] = MACHINE_LAST_POSITION[i]
            if i in line_Params:
                if line_Params[i] != H_Line_New[i] or line_M_funct != "":
                    H_Line += " " + HEIDEN_Format(i, line_Params[i])
                    H_Line_New[i] = line_Params[i]

    if H_Line == "L":  # No movements no line
        return None

    if COMPENSATION_DIFF_STATUS[0]:  # Diff from compensated ad not compensated path
        if COMPENSATION_DIFF_STATUS[1]:  # skip if already compensated, not active by now
            Cmd_Number -= 1  # align
            # initialize like true, set false if not same point compensated and not compensated
            i = True
            for j in H_Line_Params[0]:
                if j in STORED_COMPENSATED_OBJ[Cmd_Number].Parameters and j in line_Params:
                    if STORED_COMPENSATED_OBJ[Cmd_Number].Parameters[j] != line_Params[j]:
                        i = False
            if i == False:
                H_Line_Params[1][0] = "R" + line_comp
        #                   we can skip this control if already in compensation
        #                   COMPENSATION_DIFF_STATUS[1] = False
        else:
            H_Line_Params[1][0] = "R" + line_comp  # not used by now

    # check if we need to skip already active parameters
    # R parameter
    if MACHINE_SKIP_PARAMS == False or H_Line_Params[1][0] != MACHINE_STORED_PARAMS[0]:
        MACHINE_STORED_PARAMS[0] = H_Line_Params[1][0]
        H_Line += " " + H_Line_Params[1][0]

    # F parameter (check rapid o feed)
    if line_rapid:
        H_Line_Params[1][1] = FEED_MAX_SPEED
    if MACHINE_USE_FMAX and line_rapid:
        H_Line += " FMAX"
    else:
        if MACHINE_SKIP_PARAMS == False or H_Line_Params[1][1] != MACHINE_STORED_PARAMS[1]:
            MACHINE_STORED_PARAMS[1] = H_Line_Params[1][1]
            H_Line += HEIDEN_Format(" F", H_Line_Params[1][1])

    # M parameter
    if MACHINE_SKIP_PARAMS == False or H_Line_Params[1][2] != MACHINE_STORED_PARAMS[2]:
        MACHINE_STORED_PARAMS[2] = H_Line_Params[1][2]
        H_Line += " M" + H_Line_Params[1][2]

    # LBLIZE check and array creation
    if LBLIZE_STAUS:
        i = H_Line_Params[0][MACHINE_WORK_AXIS]
        # to skip reposition movements rapid or not
        if MACHINE_LAST_POSITION[i] == H_Line_New[i] and line_rapid == False:
            HEIDEN_LBL_Get(MACHINE_LAST_POSITION, H_Line_New[i])
        else:
            HEIDEN_LBL_Get()

    # update machine position with new values
    for i in H_Line_New:
        MACHINE_LAST_POSITION[i] = H_Line_New[i]

    return H_Line


# create a arc movement
def HEIDEN_Arc(arc_Params, arc_direction, arc_comp, arc_feed, arc_rapid, arc_M_funct, Cmd_Number):
    global FEED_MAX_SPEED
    global COMPENSATION_DIFF_STATUS
    global G_FUNCTION_STORE
    global MACHINE_WORK_AXIS
    global MACHINE_LAST_POSITION
    global MACHINE_LAST_CENTER
    global MACHINE_STORED_PARAMS
    global MACHINE_SKIP_PARAMS
    global MACHINE_USE_FMAX
    Cmd_Number -= 1
    H_ArcSameCenter = False
    H_ArcIncr = ""
    H_ArcCenter = "CC "
    H_ArcPoint = "C"
    H_Arc_Params = [["X", "Y", "Z"], ["R0", arc_feed, arc_M_funct], ["I", "J", "K"]]
    H_Arc_CC = {"X": 99999, "Y": 99999, "Z": 99999}  # CC initial values to overwrite
    H_Arc_P_NEW = {  # end point initial values to overwrite
        "X": 99999,
        "Y": 99999,
        "Z": 99999,
    }

    # get command values
    if G_FUNCTION_STORE["G91"]:  # incremental
        H_ArcIncr = "I"
        for i in range(0, 3):
            a = H_Arc_Params[0][i]
            b = H_Arc_Params[2][i]
            # X Y Z
            if a in arc_Params:
                H_Arc_P_NEW[a] = arc_Params[a]
            else:
                H_Arc_P_NEW[a] = 0
            # I J K skip update for machine work axis
            if i != MACHINE_WORK_AXIS:
                if b in arc_Params:
                    H_Arc_CC[a] = arc_Params[b]
                else:
                    H_Arc_CC[a] = 0
    else:  # absolute
        for i in range(0, 3):
            a = H_Arc_Params[0][i]
            b = H_Arc_Params[2][i]
            # X Y Z
            H_Arc_P_NEW[a] = MACHINE_LAST_POSITION[a]
            if a in arc_Params:
                H_Arc_P_NEW[a] = arc_Params[a]
            # I J K skip update for machine work axis
            if i != MACHINE_WORK_AXIS:
                H_Arc_CC[a] = MACHINE_LAST_POSITION[a]
                if b in arc_Params:
                    # to change if I J K are not always incremental
                    H_Arc_CC[a] = H_Arc_CC[a] + arc_Params[b]

    def Axis_Select(a, b, c, incr):
        if a in arc_Params and b in arc_Params:
            _H_ArcCenter = (
                incr + HEIDEN_Format(a, H_Arc_CC[a]) + " " + incr + HEIDEN_Format(b, H_Arc_CC[b])
            )
            if c in arc_Params and arc_Params[c] != MACHINE_LAST_POSITION[c]:
                # if there are 3 axis movements it need to be polar arc
                _H_ArcPoint = HEIDEN_PolarArc(
                    H_Arc_CC[a],
                    H_Arc_CC[b],
                    H_Arc_P_NEW[a],
                    H_Arc_P_NEW[b],
                    arc_Params[c],
                    c,
                    incr,
                )
            else:
                _H_ArcPoint = (
                    " "
                    + incr
                    + HEIDEN_Format(a, H_Arc_P_NEW[a])
                    + " "
                    + incr
                    + HEIDEN_Format(b, H_Arc_P_NEW[b])
                )
            return [_H_ArcCenter, _H_ArcPoint]
        else:
            return ["", ""]

    # set the right work plane based on tool direction
    if MACHINE_WORK_AXIS == 0:  # tool on X axis
        Axis_Result = Axis_Select("Y", "Z", "X", H_ArcIncr)
    elif MACHINE_WORK_AXIS == 1:  # tool on Y axis
        Axis_Result = Axis_Select("X", "Z", "Y", H_ArcIncr)
    elif MACHINE_WORK_AXIS == 2:  # tool on Z axis
        Axis_Result = Axis_Select("X", "Y", "Z", H_ArcIncr)
    # and fill with values
    H_ArcCenter += Axis_Result[0]
    H_ArcPoint += Axis_Result[1]

    if H_ArcCenter == "CC ":  # No movements no circle
        return None

    if arc_direction == "G2":  # set the right arc direction
        H_ArcPoint += " DR-"
    else:
        H_ArcPoint += " DR+"

    if COMPENSATION_DIFF_STATUS[0]:  # Diff from compensated ad not compensated path
        if COMPENSATION_DIFF_STATUS[1]:  # skip if already compensated
            Cmd_Number -= 1  # align
            i = True
            for j in H_Arc_Params[0]:
                if j in STORED_COMPENSATED_OBJ[Cmd_Number].Parameters and j in arc_Params:
                    if STORED_COMPENSATED_OBJ[Cmd_Number].Parameters[j] != arc_Params[j]:
                        i = False
            if i == False:
                H_Arc_Params[1][0] = "R" + arc_comp
        # COMPENSATION_DIFF_STATUS[1] = False # we can skip this control if already in compensation
        else:
            H_Arc_Params[1][0] = "R" + arc_comp  # not used by now

    # check if we need to skip already active parameters

    # R parameter
    if MACHINE_SKIP_PARAMS == False or H_Arc_Params[1][0] != MACHINE_STORED_PARAMS[0]:
        MACHINE_STORED_PARAMS[0] = H_Arc_Params[1][0]
        H_ArcPoint += " " + H_Arc_Params[1][0]

    # F parameter
    if arc_rapid:
        H_Arc_Params[1][1] = FEED_MAX_SPEED
    if MACHINE_USE_FMAX and arc_rapid:
        H_ArcPoint += " FMAX"
    else:
        if MACHINE_SKIP_PARAMS == False or H_Arc_Params[1][1] != MACHINE_STORED_PARAMS[1]:
            MACHINE_STORED_PARAMS[1] = H_Arc_Params[1][1]
            H_ArcPoint += HEIDEN_Format(" F", H_Arc_Params[1][1])

    # M parameter
    if MACHINE_SKIP_PARAMS == False or H_Arc_Params[1][2] != MACHINE_STORED_PARAMS[2]:
        MACHINE_STORED_PARAMS[2] = H_Arc_Params[1][2]
        H_ArcPoint += " M" + H_Arc_Params[1][2]

    # update values to absolute if are incremental before store
    if G_FUNCTION_STORE["G91"]:  # incremental
        for i in H_Arc_Params[0]:
            H_Arc_P_NEW[i] = MACHINE_LAST_POSITION[i] + H_Arc_P_NEW[i]
            H_Arc_CC[i] = MACHINE_LAST_POSITION[i] + H_Arc_CC[i]

    # check if we can skip CC print
    if (
        MACHINE_LAST_CENTER["X"] == H_Arc_CC["X"]
        and MACHINE_LAST_CENTER["Y"] == H_Arc_CC["Y"]
        and MACHINE_LAST_CENTER["Z"] == H_Arc_CC["Z"]
    ):
        H_ArcSameCenter = True

    # LBLIZE check and array creation
    if LBLIZE_STAUS:
        i = H_Arc_Params[0][MACHINE_WORK_AXIS]
        # to skip reposition movements
        if MACHINE_LAST_POSITION[i] == H_Arc_P_NEW[i]:
            if H_ArcSameCenter:
                HEIDEN_LBL_Get(MACHINE_LAST_POSITION, H_Arc_P_NEW[i])
            else:
                HEIDEN_LBL_Get(MACHINE_LAST_POSITION, H_Arc_P_NEW[i], 1)
        else:
            HEIDEN_LBL_Get()

    # update machine position with new values
    for i in H_Arc_Params[0]:
        MACHINE_LAST_CENTER[i] = H_Arc_CC[i]
        MACHINE_LAST_POSITION[i] = H_Arc_P_NEW[i]

    # if the circle center is already the same we don't need to print it
    if H_ArcSameCenter:
        return [H_ArcPoint]
    else:
        return [H_ArcCenter, H_ArcPoint]


def HEIDEN_PolarArc(pol_cc_X, pol_cc_Y, pol_X, pol_Y, pol_Z, pol_Axis, pol_Incr):
    pol_Angle = 0
    pol_Result = ""

    # get delta distance form point to circle center
    delta_X = pol_X - pol_cc_X
    delta_Y = pol_Y - pol_cc_Y
    # prevent undefined result of atan
    if delta_X != 0 or delta_Y != 0:
        pol_Angle = math.degrees(math.atan2(delta_X, delta_Y))

    # set the appropriate zero and direction of angle
    # with X axis zero have the Y+ direction
    if pol_Axis == "X":
        pol_Angle = 90 - pol_Angle
    # with Y axis zero have the Z+ direction
    elif pol_Axis == "Y":
        pass
    # with Z axis zero have the X+ direction
    elif pol_Axis == "Z":
        pol_Angle = 90 - pol_Angle

    # set inside +0° +360° range
    if pol_Angle > 0:
        if pol_Angle >= 360:
            pol_Angle = pol_Angle - 360
    elif pol_Angle < 0:
        pol_Angle = pol_Angle + 360
        if pol_Angle < 0:
            pol_Angle = pol_Angle + 360

    pol_Result = (
        "P" + HEIDEN_Format(" PA+", pol_Angle) + " " + pol_Incr + HEIDEN_Format(pol_Axis, pol_Z)
    )

    return pol_Result


def HEIDEN_Drill(
    drill_Obj, drill_Params, drill_Type, drill_feed
):  # create a drill cycle and movement
    global FEED_MAX_SPEED
    global MACHINE_WORK_AXIS
    global MACHINE_LAST_POSITION
    global MACHINE_USE_FMAX
    global G_FUNCTION_STORE
    global STORED_CANNED_PARAMS
    drill_Defs = {"DIST": 0, "DEPTH": 0, "INCR": 0, "DWELL": 0, "FEED": drill_feed}
    drill_Output = []
    drill_Surface = None
    drill_SafePoint = 0
    drill_StartPoint = 0

    # try to get the distance from Clearance Height and Start Depth
    if hasattr(drill_Obj, "StartDepth") and hasattr(drill_Obj.StartDepth, "Value"):
        drill_Surface = drill_Obj.StartDepth.Value

    # initialize
    if "R" in drill_Params:
        # SafePoint equals to R position
        if G_FUNCTION_STORE["G91"]:  # incremental
            drill_SafePoint = drill_Params["R"] + MACHINE_LAST_POSITION["Z"]
        else:
            drill_SafePoint = drill_Params["R"]
        # Surface equals to theoric start point of drilling
        if drill_Surface is not None and drill_SafePoint > drill_Surface:
            drill_Defs["DIST"] = drill_Surface - drill_SafePoint
            drill_StartPoint = drill_Surface
        else:
            drill_Defs["DIST"] = 0
            drill_StartPoint = drill_SafePoint

    if "Z" in drill_Params:
        if G_FUNCTION_STORE["G91"]:  # incremental
            drill_Defs["DEPTH"] = drill_SafePoint + drill_Params["Z"]
        else:
            drill_Defs["DEPTH"] = drill_Params["Z"] - drill_StartPoint

    if drill_Defs["DEPTH"] > 0:
        drill_Output.append("; WARNING START DEPTH LOWER THAN FINAL DEPTH")

    drill_Defs["INCR"] = drill_Defs["DEPTH"]
    # overwrite
    if "P" in drill_Params:
        drill_Defs["DWELL"] = drill_Params["P"]
    if "Q" in drill_Params:
        drill_Defs["INCR"] = drill_Params["Q"]

    # set the parameters for rapid movements
    if MACHINE_USE_FMAX:
        if MACHINE_SKIP_PARAMS:
            drill_Rapid = " FMAX"
        else:
            drill_Rapid = " R0 FMAX M"
    else:
        if MACHINE_SKIP_PARAMS:
            drill_Rapid = HEIDEN_Format(" F", FEED_MAX_SPEED)
        else:
            drill_Rapid = " R0" + HEIDEN_Format(" F", FEED_MAX_SPEED) + " M"

    # move to drill location
    drill_Movement = "L"
    if G_FUNCTION_STORE["G91"]:  # incremental
        # update Z value to R + actual_Z if first call
        if G_FUNCTION_STORE[drill_Type] == False:
            MACHINE_LAST_POSITION["Z"] = drill_Defs["DIST"] + MACHINE_LAST_POSITION["Z"]
            drill_Movement = "L" + HEIDEN_Format(" Z", MACHINE_LAST_POSITION["Z"]) + drill_Rapid
            drill_Output.append(drill_Movement)

        # update X and Y position
        if "X" in drill_Params and drill_Params["X"] != 0:
            MACHINE_LAST_POSITION["X"] = drill_Params["X"] + MACHINE_LAST_POSITION["X"]
            drill_Movement += HEIDEN_Format(" X", MACHINE_LAST_POSITION["X"])
        if "Y" in drill_Params and drill_Params["Y"] != 0:
            MACHINE_LAST_POSITION["Y"] = drill_Params["Y"] + MACHINE_LAST_POSITION["Y"]
            drill_Movement += HEIDEN_Format(" Y", MACHINE_LAST_POSITION["Y"])
        if drill_Movement != "L":  # same location
            drill_Output.append(drill_Movement + drill_Rapid)
    else:  # not incremental
        # check if R is higher than actual Z and move if needed
        if drill_SafePoint > MACHINE_LAST_POSITION["Z"]:
            drill_Movement = "L" + HEIDEN_Format(" Z", drill_SafePoint) + drill_Rapid
            drill_Output.append(drill_Movement)
            MACHINE_LAST_POSITION["Z"] = drill_SafePoint

        # update X and Y position
        if "X" in drill_Params and drill_Params["X"] != MACHINE_LAST_POSITION["X"]:
            MACHINE_LAST_POSITION["X"] = drill_Params["X"]
            drill_Movement += HEIDEN_Format(" X", MACHINE_LAST_POSITION["X"])
        if "Y" in drill_Params and drill_Params["Y"] != MACHINE_LAST_POSITION["Y"]:
            MACHINE_LAST_POSITION["Y"] = drill_Params["Y"]
            drill_Movement += HEIDEN_Format(" Y", MACHINE_LAST_POSITION["Y"])
        if drill_Movement != "L":  # same location
            drill_Output.append(drill_Movement + drill_Rapid)

        # check if R is not than actual Z and move if needed
        if drill_SafePoint != MACHINE_LAST_POSITION["Z"]:
            drill_Movement = "L" + HEIDEN_Format(" Z", drill_SafePoint) + drill_Rapid
            drill_Output.append(drill_Movement)
            MACHINE_LAST_POSITION["Z"] = drill_SafePoint

    # check if cycle is already stored
    i = True
    for j in drill_Defs:
        if drill_Defs[j] != STORED_CANNED_PARAMS[j]:
            i = False

    if i == False:  # not same cycle, update and print
        for j in drill_Defs:
            STORED_CANNED_PARAMS[j] = drill_Defs[j]

        # get the DEF template and replace the strings
        drill_CycleDef = MACHINE_CYCLE_DEF[1].format(
            DIST=str("{:.3f}".format(drill_Defs["DIST"])),
            DEPTH=str("{:.3f}".format(drill_Defs["DEPTH"])),
            INCR=str("{:.3f}".format(drill_Defs["INCR"])),
            DWELL=str("{:.0f}".format(drill_Defs["DWELL"])),
            FEED=str("{:.0f}".format(drill_Defs["FEED"])),
        )
        drill_Output.extend(drill_CycleDef.split("\n"))

    # add the cycle call to do the drill
    if MACHINE_SKIP_PARAMS:
        drill_Output.append("CYCL CALL")
    else:
        drill_Output.append("CYCL CALL M")

    # set already active cycle,    to do: check changes and movements until G80
    G_FUNCTION_STORE[drill_Type] = True

    return drill_Output


def HEIDEN_Format(formatType, formatValue):
    global SPINDLE_DECIMALS
    global FEED_DECIMALS
    global AXIS_DECIMALS
    returnString = ""

    formatType = str(formatType)
    if formatType == "S" or formatType == " S":
        returnString = "%.*f" % (SPINDLE_DECIMALS, formatValue)
    elif formatType == "F" or formatType == " F":
        returnString = "%.*f" % (FEED_DECIMALS, formatValue)
    else:
        returnString = "%.*f" % (AXIS_DECIMALS, formatValue)
    return formatType + str(returnString)


def HEIDEN_Numberize(GCodeStrings):  # add line numbers and concatenation
    global LINENUMBERS
    global STARTLINENR
    global LINENUMBER_INCREMENT

    if LINENUMBERS:
        linenr = STARTLINENR
        result = ""
        for s in GCodeStrings:
            result += str(linenr) + " " + s + "\n"
            linenr += LINENUMBER_INCREMENT
        return result
    else:
        return "\n".join(GCodeStrings)


def HEIDEN_LBL_Get(GetPoint=None, GetLevel=None, GetAddit=0):
    global POSTGCODE
    global STORED_LBL
    global LBLIZE_PATH_LEVEL

    if GetPoint is not None:
        if LBLIZE_PATH_LEVEL != GetLevel:
            LBLIZE_PATH_LEVEL = GetLevel
            if len(STORED_LBL) != 0 and len(STORED_LBL[-1][-1]) == 0:
                STORED_LBL[-1].pop()
            STORED_LBL.append([[len(POSTGCODE) + GetAddit, len(POSTGCODE) + GetAddit]])
        else:
            if len(STORED_LBL[-1][-1]) == 0:
                STORED_LBL[-1][-1][0] = len(POSTGCODE) + GetAddit
            STORED_LBL[-1][-1][1] = len(POSTGCODE) + GetAddit
    else:
        if len(STORED_LBL) != 0 and len(STORED_LBL[-1][-1]) != 0:
            STORED_LBL[-1].append([])
    return


def HEIDEN_LBL_Replace():
    global POSTGCODE
    global STORED_LBL
    global FIRST_LBL

    Gcode_Lenght = len(POSTGCODE)

    # this function look inside strings to find and replace it with LBL
    # the array is bi-dimensional every "Z" step down create a column
    # until the "Z" axis is moved or rapid movements are performed
    # these "planar with feed movements" create a new row with
    # the index of start and end POSTGCODE line
    # to LBLize we need to diff with the first column backward
    # from the last row in the last column of indexes
    Cols = len(STORED_LBL) - 1
    if Cols > 0:
        FIRST_LBL += len(STORED_LBL[0])  # last LBL number to print
        for Col in range(Cols, 0, -1):
            LBL_Shift = 0  # LBL number iterator
            Rows = len(STORED_LBL[0]) - 1
            for Row in range(Rows, -1, -1):
                LBL_Shift += 1
                # get the indexes from actual column
                CellStart = STORED_LBL[Col][Row][1]
                CellStop = STORED_LBL[Col][Row][0] - 1
                # get the indexes from first column
                RefStart = STORED_LBL[0][Row][1]
                i = 0
                Remove = True  # initial value True, be False on error
                for j in range(CellStart, CellStop, -1):
                    # diff from actual to first index column/row
                    if POSTGCODE[j] != POSTGCODE[RefStart - i]:
                        Remove = False
                        break
                    i += 1
                if Remove:
                    # we can remove duplicate movements
                    for j in range(CellStart, CellStop, -1):
                        POSTGCODE.pop(j)
                    # add the LBL calls
                    POSTGCODE.insert(CellStop + 1, "CALL LBL " + str(FIRST_LBL - LBL_Shift))
        if Gcode_Lenght != len(POSTGCODE):
            LBL_Shift = 0
            Rows = len(STORED_LBL[0]) - 1
            for Row in range(Rows, -1, -1):
                LBL_Shift += 1
                RefEnd = STORED_LBL[0][Row][1] + 1
                RefStart = STORED_LBL[0][Row][0]
                POSTGCODE.insert(RefEnd, "LBL 0")
                POSTGCODE.insert(RefStart, "LBL " + str(FIRST_LBL - LBL_Shift))
    STORED_LBL.clear()
    return
